"""
This module handles the topological elements of force fields.
"""


from peleffy.topology.elements import (Atom, Bond, Angle,
                                       OFFProper, OFFImproper)
from peleffy.utils.toolkits import RDKitToolkitWrapper
from peleffy.utils import Logger


class Topology(object):
    """
    It represents the topology of a molecule.
    """

    def __init__(self, molecule, parameters):
        """
        It initializes a molecular topology representation, given a
        molecule and its parameters.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object whose topology will be built
        parameters : a BaseParameterWrapper
            The parameter wrapper belonging to the molecule

        Examples
        --------

        Load a molecule from a PDB file, parameterize it with a force
        field and generate its topology

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')
        >>> molecule.parameterize('openff_unconstrained-1.2.0.offxml')

        >>> from peleffy.forcefield import OpenForceField

        >>> openff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        >>> parameters = openff.parameterize(molecule)

        >>> from peleffy.topology import Topology

        >>> topology = Topology(molecule, parameters)

        Once a topology has been built, topological elements such as
        atoms, bonds, angles and dihedrals are easily accessible
        through the attributes below

        >>> topology.atoms

        >>> topology.bonds

        >>> topology.angles

        >>> topology.propers

        >>> topology.impropers

        """

        # Set topology attributes
        self._molecule = molecule
        self._parameters = parameters

        # Initialize and build the topology
        self._initialize()
        self._build()

    def _initialize(self):
        """It initializes all the lists before building the topology."""
        self._atoms = list()
        self._bonds = list()
        self._angles = list()
        self._propers = list()
        self._OFF_propers = list()
        self._impropers = list()
        self._OFF_impropers = list()

    def _build(self):
        """The topology builder."""

        # In case the molecule has not been initialized
        if (self.molecule.rdkit_molecule is None
                or len(list(self.parameters.atom_iterator)) == 0):
            logger = Logger()
            logger.warning('Warning: the input molecule has not been '
                           + ' initialized and its topology will be empty')
            return

        self._build_atoms()
        self._build_bonds()
        self._build_angles()
        self._build_propers()
        self._build_impropers()

    def _build_atoms(self):
        """It builds the atoms of the molecule."""

        from peleffy.utils import Logger
        logger = Logger()

        coords = RDKitToolkitWrapper().get_coordinates(self.molecule)

        for index, (atom_name, atom_type, sigma, epsilon, charge,
                    SGB_radius, vdW_radius, gamma, alpha) \
                in enumerate(self.parameters.atom_iterator):
            atom = Atom(index=index,
                        PDB_name=atom_name,
                        OPLS_type=atom_type,
                        x=coords[index][0],
                        y=coords[index][1],
                        z=coords[index][2],
                        sigma=sigma,
                        epsilon=epsilon,
                        charge=charge,
                        born_radius=SGB_radius,
                        SASA_radius=vdW_radius,
                        nonpolar_gamma=gamma,
                        nonpolar_alpha=alpha)
            self.add_atom(atom)

        for atom in self.atoms:
            if atom.index in self.molecule.graph.core_nodes:
                atom.set_as_core()
            else:
                atom.set_as_branch()

        # Start from an atom from the core
        absolute_parent = None
        for atom in self.atoms:
            if atom.core:
                absolute_parent = atom.index
                break
        else:
            logger.error('Error: no core atom found in molecule '
                         + '{}'.format(self.molecule.name))

        # Get parent indexes from the molecular graph
        parent_idxs = self.molecule.graph.get_parents(absolute_parent)

        # Assert parent_idxs has right length
        if len(parent_idxs) != len(self.atoms):
            logger.error('Error: no core atom found in molecule '
                         + '{}'.format(self.molecule.name))

        for atom in self.atoms:
            parent_idx = parent_idxs[atom.index]
            if parent_idx is not None:
                atom.set_parent(self.atoms[parent_idx])

    def _build_bonds(self):
        """It builds the bonds of the molecule."""
        for index, bond in enumerate(self.parameters['bonds']):
            bond = Bond(index=index,
                        atom1_idx=bond['atom1_idx'],
                        atom2_idx=bond['atom2_idx'],
                        spring_constant=bond['spring_constant'],
                        eq_dist=bond['eq_dist'])
            self.add_bond(bond)

    def _build_angles(self):
        """It builds the angles of the molecule."""
        for index, angle in enumerate(self.parameters['angles']):
            angle = Angle(index=index,
                          atom1_idx=angle['atom1_idx'],
                          atom2_idx=angle['atom2_idx'],
                          atom3_idx=angle['atom3_idx'],
                          spring_constant=angle['spring_constant'],
                          eq_angle=angle['eq_angle'])
            self.add_angle(angle)

    def _build_propers(self):
        """It builds the propers of the molecule."""
        for index, proper in enumerate(self.parameters['propers']):
            off_proper = OFFProper(atom1_idx=proper['atom1_idx'],
                                   atom2_idx=proper['atom2_idx'],
                                   atom3_idx=proper['atom3_idx'],
                                   atom4_idx=proper['atom4_idx'],
                                   periodicity=proper['periodicity'],
                                   phase=proper['phase'],
                                   k=proper['k'],
                                   idivf=proper['idivf'])

            PELE_proper = off_proper.to_PELE()
            self.add_proper(PELE_proper)
            self.add_OFF_proper(off_proper)

        self._handle_excluded_propers()

    def _handle_excluded_propers(self):
        """
        It looks for those propers that define duplicated 1-4 relations
        and sets them to be ignored in PELE's 1-4 list.
        """
        for i, proper in enumerate(self.propers):
            atom1_idx = proper.atom1_idx
            atom4_idx = proper.atom4_idx
            for proper_to_compare in self.propers[0:i]:
                if proper == proper_to_compare:
                    continue

                if proper_to_compare.atom3_idx < 0:
                    continue

                # PELE already ignores 1-4 pair when the proper is exactly
                # the same
                if (proper.atom1_idx == proper_to_compare.atom1_idx
                        and proper.atom2_idx == proper_to_compare.atom2_idx
                        and proper.atom3_idx == proper_to_compare.atom3_idx
                        and proper.atom4_idx == proper_to_compare.atom4_idx):
                    continue

                atom1_idx_to_compare = proper_to_compare.atom1_idx
                atom4_idx_to_compare = proper_to_compare.atom4_idx
                if (atom1_idx == atom1_idx_to_compare
                        and atom4_idx == atom4_idx_to_compare):
                    proper.exclude_from_14_list()
                elif (atom1_idx == atom4_idx_to_compare
                        and atom4_idx == atom1_idx_to_compare):
                    proper.exclude_from_14_list()

            for angle_to_compare in self.angles:
                atom1_idx_to_compare = angle_to_compare.atom1_idx
                atom4_idx_to_compare = angle_to_compare.atom3_idx
                if (atom1_idx == atom1_idx_to_compare
                        and atom4_idx == atom4_idx_to_compare):
                    proper.exclude_from_14_list()
                elif (atom1_idx == atom4_idx_to_compare
                        and atom4_idx == atom1_idx_to_compare):
                    proper.exclude_from_14_list()

            for bond_to_compare in self.bonds:
                atom1_idx_to_compare = bond_to_compare.atom1_idx
                atom4_idx_to_compare = bond_to_compare.atom2_idx
                if (atom1_idx == atom1_idx_to_compare
                        and atom4_idx == atom4_idx_to_compare):
                    proper.exclude_from_14_list()
                elif (atom1_idx == atom4_idx_to_compare
                        and atom4_idx == atom1_idx_to_compare):
                    proper.exclude_from_14_list()

    def _build_impropers(self):
        """It builds the impropers of the molecule."""
        for index, improper in enumerate(self.parameters['impropers']):
            off_improper = OFFImproper(atom1_idx=improper['atom1_idx'],
                                       atom2_idx=improper['atom2_idx'],
                                       atom3_idx=improper['atom3_idx'],
                                       atom4_idx=improper['atom4_idx'],
                                       periodicity=improper['periodicity'],
                                       phase=improper['phase'],
                                       k=improper['k'],
                                       idivf=improper['idivf'])

            PELE_improper = off_improper.to_PELE()
            self.add_improper(PELE_improper)
            self.add_OFF_improper(off_improper)

    def add_atom(self, atom):
        """
        It adds an atom to the molecule's list of atoms.

        Parameters
        ----------
        atom : an peleffy.topology.Atom
            The Atom to add
        """
        self._atoms.append(atom)

    def add_bond(self, bond):
        """
        It adds a bond to the molecule's list of bonds.

        Parameters
        ----------
        bond : an peleffy.topology.Bond
            The Bond to add
        """
        self._bonds.append(bond)

    def add_angle(self, angle):
        """
        It adds an angle to the molecule's list of angles.

        Parameters
        ----------
        angle : an peleffy.topology.Angle
            The Angle to add
        """
        self._angles.append(angle)

    def add_proper(self, proper):
        """
        It adds a proper dihedral to the molecule's list of propers.

        Parameters
        ----------
        proper : an peleffy.topology.Proper
            The Proper to add
        """
        self._propers.append(proper)

    def add_OFF_proper(self, proper):
        """
        It adds a proper dihedral to the molecule's list of OFF propers.

        Parameters
        ----------
        proper : an peleffy.topology.OFFProper
            The OFFProper to add
        """
        self._OFF_propers.append(proper)

    def add_improper(self, improper):
        """
        It adds an improper dihedral to the molecule's list of impropers.

        Parameters
        ----------
        improper : an peleffy.topology.Improper
            The Improper to add
        """
        self._impropers.append(improper)

    def add_OFF_improper(self, improper):
        """
        It adds an improper dihedral to the molecule's list of OFF impropers.

        Parameters
        ----------
        improper : an peleffy.topology.OFFImproper
            The OFFImproper to add
        """
        self._OFF_impropers.append(improper)

    @property
    def molecule(self):
        """
        It returns the molecule that belongs to this topology element.

        Returns
        -------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object belonging to the topology
        """
        return self._molecule

    @property
    def parameters(self):
        """
        It returns the parameter wrapper that belongs to this topology
        element.

        Returns
        -------
        parameters : a BaseParameterWrapper
            The parameter wrapper belonging to the topology
        """
        return self._parameters

    @property
    def atoms(self):
        """
        The list of atoms in the topology.

        Returns
        -------
        atoms : list[peleffy.topology.molecule.Atom]
            The list of atoms of this Topology object.
        """
        return self._atoms

    @property
    def bonds(self):
        """
        The list of bonds in the topology.

        Returns
        -------
        bonds : list[peleffy.topology.Bond]
            The list of bonds of this Topology object.
        """
        return self._bonds

    @property
    def angles(self):
        """
        The list of angles in the topology.

        Returns
        -------
        angles : list[peleffy.topology.Angle]
            The list of angles of this Topology object.
        """
        return self._angles

    @property
    def propers(self):
        """
        The list of propers in the topology.

        Returns
        -------
        propers : list[peleffy.topology.Proper]
            The list of propers of this Topology object.
        """
        return self._propers

    @property
    def impropers(self):
        """
        The list of impropers in the topology.

        Returns
        -------
        impropers : list[peleffy.topology.Improper]
            The list of impropers of this Topology object.
        """
        return self._impropers



"""
This module contains external toolkit wrappers that are required by the
main peleffy modules.
"""

import importlib
from distutils.spawn import find_executable
import tempfile
import os
import subprocess
from pathlib import Path
from copy import deepcopy

import numpy as np
from simtk import unit

from peleffy.utils import temporary_cd


class ToolkitUnavailableException(Exception):
    """The requested toolkit is unavailable."""
    pass


class ChargeCalculationError(Exception):
    """An external error when calculating charges"""
    pass


class ChargeMethodUnavailableError(Exception):
    """A toolkit does not support the requested partial_charge_method combination"""
    pass


class ToolkitWrapper(object):
    """
    Toolkit wrapper base class.
    """

    _is_available = None
    _toolkit_name = None

    @property
    def toolkit_name(self):
        """
        The name of the toolkit.

        Returns
        -------
        toolkit_name : str
            The name of this ToolkitWrapper object
        """
        return self._toolkit_name

    @staticmethod
    def is_available():
        """
        Check whether the corresponding toolkit can be imported
        Returns
        -------
        is_installed : bool
            True if corresponding toolkit is installed, False otherwise.
        """
        return NotImplementedError


class RDKitToolkitWrapper(ToolkitWrapper):
    """
    RDKitToolkitWrapper class.
    """

    _toolkit_name = 'RDKit Toolkit'

    def __init__(self):
        """
        It initializes a RDKitToolkitWrapper object.
        """
        super().__init__()

        if not self.is_available():
            raise ToolkitUnavailableException(
                'The required toolkit {} is not '.format(self.toolkit_name)
                + 'available.')

    @staticmethod
    def is_available():
        """
        Check whether the RDKit toolkit can be imported

        Returns
        -------
        is_installed : bool
            True if RDKit is installed, False otherwise.
        """
        try:
            importlib.import_module('rdkit', 'Chem')
            return True
        except ImportError:
            return False

    def from_pdb(self, path):
        """
        It initializes an RDKit's Molecule object from a PDB file.

        Parameters
        ----------
        path : str
            The path to the molecule's PDB file

        Returns
        -------
        molecule : an rdkit.Chem.rdchem.Mol object
            The RDKit's Molecule object
        """
        from rdkit import Chem

        return Chem.rdmolfiles.MolFromPDBFile(path, removeHs=False)

    def from_pdb_block(self, pdb_block):
        """
        It initializes an RDKit's Molecule object from a PDB block.

        Parameters
        ----------
        pdb_block : str
            The PDB block to built the molecule with

        Returns
        -------
        molecule : an rdkit.Chem.rdchem.Mol object
            The RDKit's Molecule object
        """
        from rdkit import Chem

        return Chem.rdmolfiles.MolFromPDBBlock(pdb_block, removeHs=False)

    def from_smiles(self, smiles):
        """
        It initializes an RDKit's Molecule object from a SMILES tag.

        Parameters
        ----------
        smiles : str
            The SMILES tag to construct the molecule structure with.

        Returns
        -------
        molecule : an rdkit.Chem.rdchem.Mol object
            The RDKit's Molecule object
        """
        from rdkit.Chem import AllChem as Chem

        molecule = Chem.MolFromSmiles(smiles)

        # Add hydrogens to molecule
        molecule = Chem.AddHs(molecule)

        # Generate 3D coordinates
        Chem.EmbedMolecule(molecule)

        return molecule

    def assign_connectivity_from_template(self, molecule):
        """
        It assigns the connectivity to an RDKit molecule according to the
        connectivity from an RDKit connectivity template.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        """
        from rdkit.Chem import AllChem

        if molecule.connectivity_template is None:
            raise ValueError('A connectivity template must be previously '
                             + 'assigned to the molecule')

        rdkit_molecule = molecule.rdkit_molecule

        rdkit_molecule = AllChem.AssignBondOrdersFromTemplate(
            molecule.connectivity_template, rdkit_molecule)

        molecule._rdkit_molecule = rdkit_molecule

    def assign_stereochemistry_from_3D(self, molecule):
        """
        It assigns the stereochemistry to an RDKit molecule according to the
        3D coordinates in the PDB structure.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        """
        from rdkit import Chem

        rdkit_molecule = molecule.rdkit_molecule
        Chem.rdmolops.AssignStereochemistryFrom3D(rdkit_molecule)

    def set_conformer(self, molecule, conformer):
        """
        It sets a new conformation to the molecule.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        conformer : an RDKit.Chem.rdchem.Conformer object
            The conformer to set to the molecule
        """

        rdkit_molecule = molecule.rdkit_molecule

        # Remove previous conformer
        rdkit_molecule.RemoveAllConformers()

        # Add current conformer
        rdkit_molecule.AddConformer(conformer, assignId=True)

    def get_residue_name(self, molecule):
        """
        It returns the name of the residue according to the RDKit molecule
        object.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        residue_name : str
            The name of the residue
        """
        rdkit_molecule = molecule.rdkit_molecule

        first_atom = list(rdkit_molecule.GetAtoms())[0]

        # Catch a None return
        try:
            residue_name = first_atom.GetPDBResidueInfo().GetResidueName()
        except AttributeError:
            residue_name = None

        return residue_name

    def get_atom_names(self, molecule):
        """
        It returns the ordered list of atom names according to the
        RDKit molecule object. In case no atom names are available
        (non-PDB source), it assignes a name to each atom considering
        the element and an index obtained from the total number of
        occurrences of each element.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        residue_name : list[str]
            The list of atom names
        """
        rdkit_molecule = molecule.rdkit_molecule

        atom_names = list()
        occurrences = dict()

        for atom in rdkit_molecule.GetAtoms():
            pdb_info = atom.GetPDBResidueInfo()

            if pdb_info is not None:
                atom_names.append(pdb_info.GetName())
            else:
                element = atom.GetSymbol()
                occurrences[element] = occurrences.get(element, 0) + 1

                atom_names.append('{:^4}'.format(str(element)
                                                 + str(occurrences[element])))

        return atom_names

    def get_dihedral(self, mol, atom1, atom2, atom3, atom4, units="radians"):
        """
        It calculates the value of the dihedral angle in the specified units
            (default radians)

        Parameters
        ----------
        molecule : an offpele.topology.Molecule
            The offpele's Molecule object
        atom1 : int
            Index of the first atom in the dihedral
        atom2 : int
            Index of the second atom in the dihedral
        atom3 : int
            Index of the third atom in the dihedral
        atom4 : int
            Index of the fourth atom in the dihedral
        units : str
            The units in which to calculate the angle (default is radians, can
            be radians or degrees)
        """
        from rdkit.Chem import rdMolTransforms
        if units == "degrees":
            angle = rdMolTransforms.GetDihedralDeg(mol.rdkit_molecule.GetConformer(), atom1, atom2, atom3, atom4)
        else:
            angle = rdMolTransforms.GetDihedralRad(mol.rdkit_molecule.GetConformer(), atom1, atom2, atom3, atom4)
        return angle

    def get_substruct_match(self, mol1, mol2):
        """
        It returns the atoms in mol2 that match those of mol1

        Parameters
        ----------
        mol1 : an offpele.topology.Molecule
            Molecule to match atoms from
        mol2 : an offpele.topology.Molecule
            Molecule with the atoms to match

        Returns
        -------
        mol2_atoms : tuple[int]
            The tuple of atom indices from mol2 that match those in mol1
        """
        mol2_atoms = mol1.rdkit_molecule.GetSubstructMatch(mol2.rdkit_molecule)

        return mol2_atoms

    def get_atom_degrees(self, molecule):
        """
        It returns the ordered list of atom degrees. The degree of an atom
        is defined as the number of directly-bonded neighbors. Note that
        the degree is independent of bond orders.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        atom_degrees : list[int]
            The list of atom degrees
        """
        rdkit_molecule = molecule.rdkit_molecule

        atom_degrees = list()

        for atom in rdkit_molecule.GetAtoms():
            atom_degrees.append(atom.GetDegree())

        return atom_degrees

    def get_hydrogen_parents(self, molecule):
        """
        It returns the ordered list of the element belonging to the atom
        parent for the hydrogen atoms of the molecule.

        Note that this functions sets the element to None when
        the child is not a hydrogen atom.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        atom_parents : list[int]
            The list of elements belonging to the atom parent, if the
            atom is an hydrogen. The element is set to None when the
            child is not a hydrogen atom
        """
        rdkit_molecule = molecule.rdkit_molecule

        atom_parents = list()

        for atom in rdkit_molecule.GetAtoms():
            if atom.GetSymbol() == 'H':
                bonds = atom.GetBonds()

                assert len(bonds) == 1, \
                    'Hydrogen atom should only have 1 bond'

                if bonds[0].GetBeginAtom().GetSymbol() == 'H':
                    atom_parents.append(bonds[0].GetEndAtom().GetSymbol())

                else:
                    atom_parents.append(bonds[0].GetBeginAtom().GetSymbol())

            else:
                atom_parents.append(None)

        return atom_parents

    def get_elements(self, molecule):
        """
        It returns the ordered list of elements of the molecule.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        elements : list[str]
            The list of elements belonging to supplied Molecule object
        """
        rdkit_molecule = molecule.rdkit_molecule

        elements = list()

        for atom in rdkit_molecule.GetAtoms():
            elements.append(atom.GetSymbol())

        return elements

    def to_pdb_file(self, molecule, path):
        """
        It writes the RDKit molecule to a PDB file.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        path : str
            Path to write to
        """
        from rdkit import Chem

        assert Path(path).suffix == '.pdb', 'Wrong extension'

        rdkit_molecule = molecule.rdkit_molecule

        pdb_block = Chem.rdmolfiles.MolToPDBBlock(rdkit_molecule)
        names = molecule.get_pdb_atom_names()
        tag = molecule.tag

        renamed_pdb_block = ''
        atom_counter = 0
        for line in pdb_block.split('\n'):
            if line.startswith('HETATM'):
                renamed_pdb_block += line[:12] + names[atom_counter] \
                    + ' ' + tag + line[20:] + '\n'
                atom_counter += 1
            else:
                renamed_pdb_block += line + '\n'

        with open(path, 'w') as f:
            f.write(renamed_pdb_block)

    def to_sdf_file(self, molecule, path):
        """
        It writes the RDKit molecule to an sdf file.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        path : str
            Path to write to
        """
        from rdkit import Chem

        assert Path(path).suffix == '.sdf', 'Wrong extension'

        rdkit_molecule = molecule.rdkit_molecule
        with open(path, 'w') as f:
            writer = Chem.SDWriter(f)
            writer.write(rdkit_molecule)
            writer.close()

    def to_xyz_file(self, molecule, path):
        """
        It writes the RDKit molecule to an xyz file.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        path : str
            Path to write to
        """
        from rdkit import Chem

        assert Path(path).suffix == '.xyz', 'Wrong extension'

        rdkit_molecule = molecule.rdkit_molecule
        Chem.MolToXYZFile(rdkit_molecule, path)

    def get_atom_ids_with_rotatable_bonds(self, molecule):
        """
        It returns the atom ids with rotatable bonds according to the
        RDKit molecule.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        rot_bonds_atom_ids : tuple[tuple[int, int]]
            The set of atom id pairs that belong to rotatable bonds
        """
        from rdkit import Chem

        rdkit_molecule = deepcopy(molecule.rdkit_molecule)

        rot_bonds_atom_ids = set([
            frozenset(atom_pair) for atom_pair in
            rdkit_molecule.GetSubstructMatches(
                Chem.MolFromSmarts('[!$([NH]!@C(=O))&!D1&!$(*#*)]-&!@[!$([NH]!@C(=O))&!D1&!$(*#*)]'))])

        # Include missing rotatable bonds for amide groups
        for atom_pair in [frozenset(atom_pair) for atom_pair in
                          rdkit_molecule.GetSubstructMatches(
                          Chem.MolFromSmarts('[$(N!@C(=O))]-&!@[!$(C(=O))&!D1&!$(*#*)]'))]:
            rot_bonds_atom_ids.add(atom_pair)

        # Remove bonds to terminal -CH3
        if molecule.exclude_terminal_rotamers:
            terminal_bonds = set([
                frozenset(atom_pair) for atom_pair in
                rdkit_molecule.GetSubstructMatches(
                    Chem.MolFromSmarts('*-&!@[$([C;H3;X4]),$([N;H2;X3]),$([N;H3;X4]),$([O;H1;X2])]'))
            ])
            rot_bonds_atom_ids = rot_bonds_atom_ids.difference(terminal_bonds)

        return list(rot_bonds_atom_ids)

    def get_coordinates(self, molecule):
        """
        It returns the 3D coordinates of all atoms in the RDKit molecule.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        coordinates : numpy.ndarray
            The array of 3D coordinates of all the atoms in the molecule
        """
        rdkit_molecule = molecule.rdkit_molecule

        conformer = rdkit_molecule.GetConformer()
        return conformer.GetPositions()

    def get_2D_representation(self, molecule):
        """
        It returns the 2D representation of the RDKit molecule.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        representation_2D : an RDKit.molecule object
            It is an RDKit molecule with an embeded 2D representation
        """
        from rdkit.Chem import AllChem

        rdkit_molecule = molecule.rdkit_molecule
        representation_2D = deepcopy(rdkit_molecule)

        AllChem.Compute2DCoords(representation_2D)
        return representation_2D

    def get_rmsd(self, molecule, molecule_2):
        """
        It returns the RMSD between two RDKit molecules.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        molecule_2 : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        rmsd_value : float
            RMSD between two RDKit molecules
        """
        from rdkit.Chem import rdMolAlign

        rmsd_value = rdMolAlign.AlignMol(molecule.rdkit_molecule,
                                         molecule_2.rdkit_molecule)
        return rmsd_value

    def draw_molecule(self, representation, atom_indexes=list(),
                      radii_dict=dict(), atom_color_dict=dict(),
                      bond_indexes=list(), bond_color_dict=dict()):
        """
        Given a molecular representation, it returns its image as SVG.

        Parameters
        ----------
        representation : an RDKit.molecule object
            It is an RDKit molecule with an embeded 2D representation

        Returns
        -------
        image : an IPython's SVG display object
            The image of the molecular representation to display
        """
        from rdkit.Chem.Draw import rdMolDraw2D

        draw = rdMolDraw2D.MolDraw2DSVG(500, 500)
        draw.SetLineWidth(4)
        rdMolDraw2D.PrepareAndDrawMolecule(draw, representation,
                                           highlightAtoms=atom_indexes,
                                           highlightAtomRadii=radii_dict,
                                           highlightAtomColors=atom_color_dict,
                                           highlightBonds=bond_indexes,
                                           highlightBondColors=bond_color_dict)
        draw.FinishDrawing()

        from IPython.display import SVG

        image = SVG(draw.GetDrawingText())

        return image


class AmberToolkitWrapper(ToolkitWrapper):
    """
    AmberToolkitWrapper class.
    """

    _toolkit_name = 'Amber Toolkit'

    def __init__(self):
        """
        It initializes a AmberToolkitWrapper object.
        """
        super().__init__()

        if not self.is_available():
            raise ToolkitUnavailableException(
                'The required toolkit {} is not '.format(self.toolkit_name)
                + 'available.')

        self._rdkit_toolkit_wrapper = RDKitToolkitWrapper()

    @staticmethod
    def is_available():
        """
        Check whether the AmberTools toolkit is installed

        Returns
        -------
        is_installed : bool
            True if AmberTools is installed, False otherwise.
        """
        ANTECHAMBER_PATH = find_executable("antechamber")
        if ANTECHAMBER_PATH is None:
            return False
        if not(RDKitToolkitWrapper.is_available()):
            return False
        return True

    def compute_partial_charges(self, molecule, method='am1bcc'):
        """
        It computes the partial charges using antechamber.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object
        method : str
            The name of the method to use. One of ['gasteiger', 'am1bcc'].
            If None, 'am1bcc' will be used

        Returns
        -------
        charges : simtk.unit.Quantity
            The array of partial charges

        Raises
        ------
        ChargeMethodUnavailableError if the requested charge method can not
            be handled by this toolkit
        ChargeCalculationError if the charge method is supported by this
            toolkit, but fails
        """

        SUPPORTED_CHARGE_METHODS = {'am1bcc': {'antechamber_keyword': 'bcc'},
                                    'gasteiger': {'antechamber_keyword': 'gas'}
                                    }

        if method not in SUPPORTED_CHARGE_METHODS:
            raise ChargeMethodUnavailableError(
                'partial_charge_method '
                + '{} is not available from '.format(method)
                + 'AmberToolsToolkitWrapper. Available charge methods are '
                + list(SUPPORTED_CHARGE_METHODS.keys()))

        off_molecule = molecule.off_molecule

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                net_charge = off_molecule.total_charge / \
                    unit.elementary_charge

                self._rdkit_toolkit_wrapper.to_sdf_file(
                    molecule, tmpdir + '/molecule.sdf')

                subprocess.check_output([
                    "antechamber", "-i", "molecule.sdf", "-fi", "sdf",
                    "-o", "charged.ac", "-fo", "ac", "-pf", "yes", "-dr", "n",
                    "-c",
                    SUPPORTED_CHARGE_METHODS[method]['antechamber_keyword'],
                    "-nc", str(net_charge)])
                # Write out just charges
                subprocess.check_output([
                    "antechamber", "-dr", "n", "-i", "charged.ac", "-fi", "ac",
                    "-o", "charged2.ac", "-fo", "ac", "-c", "wc",
                    "-cf", "charges.txt", "-pf", "yes"])

                if not os.path.exists('charges.txt'):
                    # TODO: copy files into local directory to aid debugging?
                    raise ChargeCalculationError(
                        "Antechamber/sqm partial charge calculation failed on "
                        "molecule {} (SMILES {})".format(
                            off_molecule.name, off_molecule.to_smiles()))

                # Read the charges
                with open('charges.txt', 'r') as infile:
                    contents = infile.read()

                text_charges = contents.split()
                charges = np.zeros([off_molecule.n_atoms], np.float64)
                for index, token in enumerate(text_charges):
                    charges[index] = float(token)

        charges = unit.Quantity(charges, unit.elementary_charge)

        assert len(charges) == len(molecule.rdkit_molecule.GetAtoms()), \
            'Partial charge computation failed as the length of ' \
            + 'resulting partial charges does not match with the ' \
            + 'number of atoms in molecule'

        return charges


class OpenForceFieldToolkitWrapper(ToolkitWrapper):
    """
    OpenForceFieldToolkitWrapper class.
    """

    _toolkit_name = 'OpenForceField Toolkit'

    def __init__(self):
        """
        It initializes a OpenForceFieldToolkitWrapper object.
        """
        super().__init__()

        if not self.is_available():
            raise ToolkitUnavailableException(
                'The required toolkit {} is not '.format(self.toolkit_name)
                + 'available.')

    @staticmethod
    def is_available():
        """
        Check whether the OpenForceField toolkit is installed

        Returns
        -------
        is_installed : bool
            True if OpenForceField is installed, False otherwise.
        """
        try:
            importlib.import_module('openforcefield')
            return True
        except ImportError:
            return False

    def from_rdkit(self, molecule):
        """
        It initializes an OpenForceField's Molecule object from an RDKit
        molecule.

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        molecule : an openforcefield.topology.Molecule object
            The OpenForceField's Molecule
        """
        from openforcefield.topology.molecule import Molecule

        rdkit_molecule = molecule.rdkit_molecule
        return Molecule.from_rdkit(
            rdkit_molecule,
            allow_undefined_stereo=molecule.allow_undefined_stereo)

    def get_forcefield(self, forcefield_name):
        """
        It returns the OpenForceField's object that matches with the name
        that is supplied.

        Parameters
        ----------
        forcefield_name : str
            The name of the requested forcefield

        Returns
        -------
        forcefield : an openforcefield.typing.engines.smirnoff.ForceField
                     object
            The OpenForceField's forcefield
        """
        from openforcefield.typing.engines.smirnoff import ForceField

        if isinstance(forcefield_name, str):
            forcefield = ForceField(forcefield_name)
        else:
            raise Exception('Invalid forcefield type')

        return forcefield

    def get_parameters_from_forcefield(self, forcefield, molecule):
        """
        It returns the parameters that are obtained with the supplied
        forcefield for a certain peleffy's molecule.

        Parameters
        ----------
        forcefield : str or an openforcefield.typing.engines.smirnoff.ForceField
                     object
            The forcefield from which the parameters will be obtained
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        openforcefield_parameters : dict
            The OpenFF parameters stored in a dict keyed by parameter type
        """
        from openforcefield.typing.engines.smirnoff import ForceField
        from openforcefield.topology import Topology

        off_molecule = molecule.off_molecule
        topology = Topology.from_molecules([off_molecule])

        if isinstance(forcefield, str):
            forcefield = ForceField(forcefield)
        elif isinstance(forcefield, ForceField):
            pass
        else:
            raise Exception('Invalid forcefield type')

        molecule_parameters_list = forcefield.label_molecules(topology)

        assert len(molecule_parameters_list) == 1, 'A single molecule is ' \
            'expected'

        return molecule_parameters_list[0]

    def get_parameter_handler_from_forcefield(self, parameter_handler_name,
                                              forcefield):
        """
        It returns a parameter handler from the forcefield based on its
        name.

        Parameters
        ----------
        parameter_handler_name : str
            The name of the parameter handler that is requested
        forcefield : an openforcefield.typing.engines.smirnoff.ForceField
                     object
            The forcefield from which the parameter handler will be obtained

        Returns
        -------
        parameter_handler : an openforcefield.typing.engines.smirnoff.parameters.ParameterHandler
                            object
            The ParameterHandler that was requested
        """
        from openforcefield.typing.engines.smirnoff import ForceField

        if isinstance(forcefield, str):
            forcefield = ForceField(forcefield)
        elif isinstance(forcefield, ForceField):
            pass
        else:
            raise Exception('Invalid forcefield type')

        return forcefield.get_parameter_handler(parameter_handler_name)


class SchrodingerToolkitWrapper(ToolkitWrapper):
    """
    SchrodingerToolkitWrapper class.
    """

    _toolkit_name = 'Schrodinger Toolkit'

    def __init__(self):
        """
        It initializes a SchrodingerToolkitWrapper object.
        """
        super().__init__()

        if "SCHRODINGER" not in os.environ:
            import logging
            logging.warning("Schrodinger Toolkit requires the environment "
                            + "variable SCHRODINGER to be previously set, "
                            + "pointing to the Schrodinger's installation "
                            + "path. For more information, please, refer to "
                            + "https://martimunicoy.github.io/peleffy/installation.html#external-dependencies",
                            )

        if not self.is_available():
            raise ToolkitUnavailableException(
                'The required toolkit {} is not '.format(self.toolkit_name)
                + 'available.')

        self._rdkit_toolkit_wrapper = RDKitToolkitWrapper()

    @staticmethod
    def is_available():
        """
        Check whether the OpenForceField toolkit is installed

        Returns
        -------
        is_installed : bool
            True if OpenForceField is installed, False otherwise.
        """
        if not(RDKitToolkitWrapper.is_available()):
            return False

        if SchrodingerToolkitWrapper.path_to_ffld_server() is None:
            return False

        return True

    @staticmethod
    def path_to_ffld_server():
        FFLD_SERVER_PATH = find_executable("ffld_server")

        if FFLD_SERVER_PATH is not None:
            return FFLD_SERVER_PATH

        else:
            if "SCHRODINGER" in os.environ:
                schrodinger_root = os.environ.get('SCHRODINGER')
                return os.path.join(schrodinger_root,
                                    'utilities', 'ffld_server')

        return None

    def run_ffld_server(self, molecule):
        """
        It calls Schrodinger's ffld_server to parameterize a molecule
        with OPLS.

        .. todo ::

           * Review PlopRotTemp's atom type fixes. Should we apply them here?

        Parameters
        ----------
        molecule : an peleffy.topology.Molecule
            The peleffy's Molecule object

        Returns
        -------
        ffld_output : str
            The ffld_server output
        """

        ffld_server_exec = self.path_to_ffld_server()

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):

                self._rdkit_toolkit_wrapper.to_pdb_file(
                    molecule, tmpdir + '/molecule.pdb')

                subprocess.check_output([ffld_server_exec,
                                         "-ipdb", "molecule.pdb",
                                         "-version", "14",
                                         "-print_parameters",
                                         "-out_file", "parameters.txt"])

                with open('parameters.txt') as parameters_file:
                    return parameters_file.read()



"""
This module contains a set of classes and methods designed to handle
output workflows.
"""


import os

from peleffy.utils import create_path


class OutputPathHandler(object):
    """
    It handles the output paths of peleffy parameter files.
    """

    OFF_IMPACT_TEMPLATE_PATH = 'DataLocal/Templates/OpenFF/Parsley/'
    OPLS_IMPACT_TEMPLATE_PATH = 'DataLocal/Templates/OPLS2005/HeteroAtoms/'
    ROTAMER_LIBRARY_PATH = 'DataLocal/LigandRotamerLibs/'
    SOLVENT_TEMPLATE_PATH = 'DataLocal/OBC/'
    CONFORMATION_LIBRARY_PATH = 'DataLocal/Conformations/'
    FILE_TYPES = ['impact template', 'rotamer library', 'solvent template',
                  'conformation library']

    def __init__(self, molecule, forcefield,
                 as_datalocal=False, output_path=None):
        """
        It initializes an OutputPathHandler object.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            A Molecule object to be written as an Impact file
        forcefield : a peleffy.forcefield.forcefield.ForceField
            The force field employed in the molecular parameterization
        as_datalocal : bool
            Whether to save output files following PELE's DataLocal
            hierarchy or not
        output_path : str
            The output path inside which all files will be saved. Default
            is the current directory
        """
        self._molecule = molecule
        self._forcefield = forcefield
        self._as_datalocal = as_datalocal
        if output_path is None:
            self._output_path = os.path.curdir
        else:
            self._output_path = output_path

    def set_DataLocal_behaviour(self, as_datalocal):
        """
        It sets the DataLocal behaviour of this OutputPathHandler object.

        Parameters
        ----------
        as_datalocal : bool
            Whether to save output files following PELE's DataLocal
            hierarchy or not
        """
        self._as_datalocal = as_datalocal

    def get_path(self, file_type, create_missing_folders=True):
        """
        It returns the path to a file, according to the file type.

        Parameters
        ----------
        file_type : str
            The file type whose path is requested. One of
            ['impact template', 'rotamer library', 'solvent template',
            'conformation library']
        create_missing_folders : bool
            Whether to create missing folders or not. Default is True

        Returns
        -------
        file_path : str
            The path for the requested file type
        """
        if file_type.lower() not in self.FILE_TYPES:
            raise ValueError('Invalid file type: {}'.format(file_type)
                             + ', it must be one of', self.FILE_TYPES)

        if file_type.lower() == 'impact template':
            return self.get_impact_template_path(create_missing_folders)

        if file_type.lower() == 'rotamer library':
            return self.get_rotamer_library_path(create_missing_folders)

        if file_type.lower() == 'solvent template':
            return self.get_solvent_template_path(create_missing_folders)

        if file_type.lower() == 'conformation library':
            return self.get_conformation_library_path(create_missing_folders)

    def get_impact_template_path(self, create_missing_folders=True):
        """
        It returns the path for an Impact template file.

        Parameters
        ----------
        create_missing_folders : bool
            Whether to create missing folders or not. Default is True

        Returns
        -------
        file_path : str
            The path for an Impact template file
        """
        file_name = self._molecule.tag.lower().strip() + 'z'

        if self.as_datalocal:
            if self.forcefield.type == 'OpenFF':
                path = os.path.join(self.output_path,
                                    self.OFF_IMPACT_TEMPLATE_PATH)
            elif self.forcefield.type == 'OpenFF + OPLS2005':
                if self.forcefield._nonbonding == 'openff':
                    path = os.path.join(self.output_path,
                                        self.OFF_IMPACT_TEMPLATE_PATH)
                else:
                    path = os.path.join(self.output_path,
                                        self.OPLS_IMPACT_TEMPLATE_PATH)
            else:
                path = os.path.join(self.output_path,
                                    self.OPLS_IMPACT_TEMPLATE_PATH)
        else:
            path = self.output_path

        if create_missing_folders:
            create_path(path)

        return os.path.join(path, file_name)

    def get_rotamer_library_path(self, create_missing_folders=True):
        """
        It returns the path for a rotamer library file.

        Parameters
        ----------
        create_missing_folders : bool
            Whether to create missing folders or not. Default is True

        Returns
        -------
        file_path : str
            The path for a rotamer library file
        """
        file_name = self._molecule.tag.upper().strip() + '.rot.assign'

        if self.as_datalocal:
            path = os.path.join(self.output_path, self.ROTAMER_LIBRARY_PATH)
        else:
            path = self.output_path

        if create_missing_folders:
            create_path(path)

        return os.path.join(path, file_name)

    def get_solvent_template_path(self, create_missing_folders=True):
        """
        It returns the path for a solvent template file.

        Parameters
        ----------
        create_missing_folders : bool
            Whether to create missing folders or not. Default is True

        Returns
        -------
        file_path : str
            The path for a solvent template file
        """
        file_name = 'ligandParams.txt'

        if self.as_datalocal:
            path = os.path.join(self.output_path, self.SOLVENT_TEMPLATE_PATH)
        else:
            path = self.output_path

        if create_missing_folders:
            create_path(path)

        return os.path.join(path, file_name)

    def get_conformation_library_path(self, create_missing_folders=True):
        """
        It returns the path for a conformation library file.

        Parameters
        ----------
        create_missing_folders : bool
            Whether to create missing folders or not. Default is True

        Returns
        -------
        file_path : str
            The path for a conformation library file
        """
        file_name = self._molecule.tag.upper() + '.conformation'

        if self.as_datalocal:
            path = os.path.join(self.output_path,
                                self.CONFORMATION_LIBRARY_PATH)
        else:
            path = self.output_path

        if create_missing_folders:
            create_path(path)

        return os.path.join(path, file_name)

    @property
    def as_datalocal(self):
        """
        The DataLocal configuration of this OutputPathHandler object.

        Returns
        -------
        as_datalocal : bool
            Whether to save output files following PELE's DataLocal
            hierarchy or not
        """
        return self._as_datalocal

    @property
    def molecule(self):
        """
        The peleffy's Molecule.

        Returns
        -------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object
        """
        return self._molecule

    @property
    def forcefield(self):
        """
        The peleffy's ForceField.

        Returns
        -------
        forcefield : a peleffy.forcefield.forcefield.ForceField
            The force field employed in the molecular parameterization
        """
        return self._forcefield

    @property
    def output_path(self):
        """
        The output path inside which all files will be saved.

        Returns
        output_path : str
            The output path
        """
        return self._output_path



"""
This module contains the tests to check peleffy's molecular representations.
"""

import pytest

import tempfile

from peleffy.topology import Molecule
from peleffy.utils import get_data_file_path, temporary_cd


class TestMolecule(object):
    """
    It wraps all tests that involve the Molecule class.
    """

    def test_pdb_initialization(self):
        """
        It checks the initialization from a PDB file.
        """
        ligand_path = get_data_file_path('ligands/ethylene.pdb')

        molecule = Molecule(ligand_path)

        # Save it
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                molecule.to_pdb_file('molecule.pdb')

    def test_smiles_initialization(self):
        """
        It checks the initialization from a SMILES tag.
        """
        molecule = Molecule(smiles='c1ccccc1')

        # Save it
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                molecule.to_pdb_file('molecule.pdb')

    def test_molecule_name_assignment(self):
        """
        It tests the molecule name assignment.
        """
        # Look for an empty name when dummy Molecule is loaded
        molecule = Molecule()
        assert molecule.name == '', 'Unexpected atom name'

        # Look for the PDB name when a Molecule is loaded from a PDB file
        ligand_path = get_data_file_path('ligands/benzene.pdb')
        molecule = Molecule(ligand_path)
        assert molecule.name == 'benzene', 'Unexpected atom name'

        # Look for benzene name when a Molecule is loaded from a PDB file
        # with a custom name
        ligand_path = get_data_file_path('ligands/benzene.pdb')
        molecule = Molecule(ligand_path, name='benzene')
        assert molecule.name == 'benzene', 'Unexpected atom name'

        # Look for the SMILES name when a Molecule is loaded from a SMILES tag
        molecule = Molecule(smiles='c1ccccc1')
        assert molecule.name == 'c1ccccc1', 'Unexpected atom name'

        # Look for benzene name when a Molecule is loaded from a SMILES tag
        # with a custom name
        molecule = Molecule(smiles='c1ccccc1', name='benzene')
        assert molecule.name == 'benzene', 'Unexpected atom name'

    def test_molecule_tag_assignment(self):
        """
        It tests the molecule tag assignment.
        """
        # Look for UNK tag when dummy Molecule is loaded
        molecule = Molecule()
        assert molecule.tag == 'UNK', 'Unexpected atom tag'

        # Look for the PDB residue name as a tag when a Molecule is loaded
        # from a PDB file
        ligand_path = get_data_file_path('ligands/benzene.pdb')
        molecule = Molecule(ligand_path)
        assert molecule.tag == 'BNZ', 'Unexpected atom tag'

        # Look for BEN tag when a Molecule is loaded from a PDB file with
        # a custom name
        ligand_path = get_data_file_path('ligands/benzene.pdb')
        molecule = Molecule(ligand_path, tag='BEN')
        assert molecule.tag == 'BEN', 'Unexpected atom tag'

        # Look for UNK tag when a Molecule is loaded from a SMILES tag
        molecule = Molecule(smiles='c1ccccc1')
        assert molecule.tag == 'UNK', 'Unexpected atom tag'

        # Look for BNZ tag when a Molecule is loaded from a SMILES tag with
        # a custom tag
        molecule = Molecule(smiles='c1ccccc1', tag='BNZ')
        assert molecule.tag == 'BNZ', 'Unexpected atom tag'

    def test_PDB_connectivity_template(self):
        """
        It tests the initialization of an peleffy's Molecule representation
        from a PDB file without connectivity and a connectivity template.
        """
        # Initialize an empty Molecule object
        molecule = Molecule()
        assert molecule.connectivity_template is None, \
            'Unexpected connectivity template'

        # Initialize a Molecule from a PDB without connectivity and
        # without a connectivity template
        ligand_path = get_data_file_path(
            'ligands/benzene_without_connectivity.pdb')
        molecule = Molecule(ligand_path)

        expected_bond_ids = [(1, 0, False), (2, 1, False), (3, 2, False),
                             (4, 3, False), (5, 4, False), (5, 0, False),
                             (6, 0, False), (7, 1, False), (8, 2, False),
                             (9, 3, False), (10, 4, False), (11, 5, False)]

        for bond in molecule.rdkit_molecule.GetBonds():
            bond_id = (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(),
                       bond.GetIsAromatic())
            assert bond_id in expected_bond_ids, 'Unexpected bond id ' \
                + '{}'.format(bond_id)

        # Initialize a Molecule from a PDB without connectivity but with
        # a connectivity template
        template_path = get_data_file_path(
            'ligands/benzene.pdb')
        template = Molecule(template_path)
        ligand_path = get_data_file_path(
            'ligands/benzene_without_connectivity.pdb')
        molecule = Molecule(ligand_path,
                            connectivity_template=template.rdkit_molecule)

        expected_bond_ids = [(1, 0, True), (2, 1, True), (3, 2, True),
                             (4, 3, True), (5, 4, True), (5, 0, True),
                             (6, 0, False), (7, 1, False), (8, 2, False),
                             (9, 3, False), (10, 4, False), (11, 5, False)]

        for bond in molecule.rdkit_molecule.GetBonds():
            bond_id = (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(),
                       bond.GetIsAromatic())
            assert bond_id in expected_bond_ids, 'Unexpected bond id ' \
                + '{}'.format(bond_id)

        # Initialize a Molecule from a PDB with connectivity and with
        # a connectivity template
        template_path = get_data_file_path(
            'ligands/benzene.pdb')
        template = Molecule(template_path)
        ligand_path = get_data_file_path(
            'ligands/benzene.pdb')
        molecule = Molecule(ligand_path,
                            connectivity_template=template.rdkit_molecule)

        expected_bond_ids = [(0, 1, True), (1, 2, True), (2, 3, True),
                             (3, 4, True), (4, 5, True), (0, 5, True),
                             (0, 6, False), (1, 7, False), (2, 8, False),
                             (3, 9, False), (4, 10, False), (5, 11, False)]

        for bond in molecule.rdkit_molecule.GetBonds():
            bond_id = (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(),
                       bond.GetIsAromatic())
            assert bond_id in expected_bond_ids, 'Unexpected bond id ' \
                + '{}'.format(bond_id)

    def test_PDB_residue_name(self):
        """
        It tests the PDB residue name and checks for consistency with
        Molecule tag.
        """

        def check_residue_name(name):
            """Check if residue names are valid in the output PDB file"""
            with open('molecule.pdb') as f:
                for line in f:
                    if line.startswith('HETATM'):
                        assert line[17:20] == name, 'Unexpected residue name'

        ligand_path = get_data_file_path('ligands/benzene.pdb')

        # Checking tag assignation from PDB
        molecule = Molecule(ligand_path)
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                assert molecule.tag == 'BNZ', 'Unexpected molecule tag'
                molecule.to_pdb_file('molecule.pdb')
                check_residue_name('BNZ')

        # Checking set_tag() function
        molecule = Molecule(ligand_path)
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                molecule.set_tag('TAG')
                assert molecule.tag == 'TAG', 'Unexpected molecule tag'
                molecule.to_pdb_file('molecule.pdb')
                check_residue_name('TAG')

        # Checking default tag assignment from SMILES
        molecule = Molecule(smiles='c1ccccc1')
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                assert molecule.tag == 'UNK', 'Unexpected molecule tag'
                molecule.to_pdb_file('molecule.pdb')
                check_residue_name('UNK')

        # Checking custom tag assignment from SMILES
        molecule = Molecule(smiles='c1ccccc1', tag='BEN')
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                assert molecule.tag == 'BEN', 'Unexpected molecule tag'
                molecule.to_pdb_file('molecule.pdb')
                check_residue_name('BEN')

        # Checking second custom tag assignment from SMILES
        molecule = Molecule(smiles='c1ccccc1')
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                molecule.set_tag('BNZ')
                assert molecule.tag == 'BNZ', 'Unexpected molecule tag'
                molecule.to_pdb_file('molecule.pdb')
                check_residue_name('BNZ')

    def test_pdb_checkup(self):
        """It tests the safety check function for PDB files."""

        LIGAND_GOOD = get_data_file_path('ligands/ethylene.pdb')
        LIGAND_ERROR1 = get_data_file_path('tests/ethylene_error1.pdb')
        LIGAND_ERROR2 = get_data_file_path('tests/ethylene_error2.pdb')
        LIGAND_ERROR3 = get_data_file_path('tests/ethylene_error3.pdb')
        LIGAND_ERROR4 = get_data_file_path('tests/ethylene_error4.pdb')

        # This should work without any complain
        _ = Molecule(LIGAND_GOOD)

        # All atom names need to be unique
        with pytest.raises(Exception):
            _ = Molecule(LIGAND_ERROR1)

        # All residue ids must match
        with pytest.raises(Exception):
            _ = Molecule(LIGAND_ERROR2)

        # All residue names must match
        with pytest.raises(Exception):
            _ = Molecule(LIGAND_ERROR3)

        # Check warning message in the logger when connectivity is missing
        import io
        from peleffy.utils import Logger
        import logging
        from importlib import reload
        logging.shutdown()
        reload(logging)

        log = Logger()
        log.set_level('WARNING')

        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            _ = Molecule(LIGAND_ERROR4)

            output = buf.getvalue()

            assert output == "Warning: input PDB has no information " \
                + "about the connectivity and this could result in " \
                + "an unexpected bond assignment\n"

    def test_undefined_stereo(self):
        """
        It checks the behaviour when ignoring the stereochemistry
        in the Molecule initialization.
        """
        from openforcefield.utils.toolkits import UndefinedStereochemistryError
        from peleffy.forcefield import OpenForceField

        # This should crash due to an undefined stereochemistry error
        with pytest.raises(UndefinedStereochemistryError):
            mol = Molecule(smiles='CN(C)CCC=C1c2ccccc2CCc3c1cccc3')

        # This now should work
        mol = Molecule(smiles='CN(C)CCC=C1c2ccccc2CCc3c1cccc3',
                       allow_undefined_stereo=True)

        # And we can parameterize it
        ff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        ff.parameterize(mol, charge_method='gasteiger')

        # Save it
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                mol.to_pdb_file('molecule.pdb')

    def test_from_rdkit(self):
        """
        It checks the initialization of a peleffy Molecule from an RDKit
        molecular representation.
        """
        from rdkit import Chem

        pdb_path = get_data_file_path('ligands/malonate.pdb')
        rdkit_molecule = Chem.MolFromPDBFile(pdb_path, removeHs=False)
        molecule = Molecule.from_rdkit(rdkit_molecule)

        assert molecule._rdkit_molecule is not None, \
            'Unexpected molecule representation found, it is not initialized'
        assert molecule._off_molecule is not None, \
            'Unexpected molecule representation found, it is not initialized'

        ref_pdb_atom_names = [' O1 ', ' C1 ', ' O2 ', ' C2 ', ' C3 ',
                              ' O3 ', ' O4 ', ' H1 ', ' H2 ', ' H3 ']
        pdb_atom_names = molecule.get_pdb_atom_names()

        assert pdb_atom_names == ref_pdb_atom_names, \
            'Unexpected PDB atom names found in the resulting Molecule ' \
            + 'representation'

        assert molecule.graph is not None, \
            'Molecule\' graph should be initialized'

    def test_from_openff(self):
        """
        It checks the initialization of a peleffy Molecule from an OpenFF
        molecular representation.
        """
        from openforcefield.topology import Molecule as OpenFFMolecule

        openff_molecule = OpenFFMolecule.from_smiles('C(C(=O)[O-])C(=O)[OH]')

        molecule = Molecule.from_openff(openff_molecule)

        assert molecule._rdkit_molecule is not None, \
            'Unexpected molecule representation found, it is not initialized'
        assert molecule._off_molecule is not None, \
            'Unexpected molecule representation found, it is not initialized'

        ref_pdb_atom_names = [' C1 ', ' C2 ', ' O1 ', ' O2 ', ' C3 ',
                              ' O3 ', ' O4 ', ' H1 ', ' H2 ', ' H3 ']
        pdb_atom_names = molecule.get_pdb_atom_names()

        assert pdb_atom_names == ref_pdb_atom_names, \
            'Unexpected PDB atom names found in the resulting Molecule ' \
            + 'representation'

        assert molecule.graph is not None, \
            'Molecule\' graph should be initialized'

    def test_pdb_fixer(self):
        """
        It checks the PDB fixer prior parsing a PDB input file for
        a peleffy Molecule.
        """
        from .utils import compare_blocks

        # Check default
        molecule = Molecule()

        assert molecule.fix_pdb is True, \
            'Unexpected default settings for the PDB fixer'

        # Activate fixer
        molecule = Molecule(fix_pdb=True)

        ref_path = get_data_file_path('tests/ligSUV_fixed.pdb')
        path1 = get_data_file_path('tests/ligSUV_no_elements1.pdb')
        path2 = get_data_file_path('tests/ligSUV_no_elements2.pdb')

        ref_pdb_block = molecule._read_and_fix_pdb(ref_path)
        pdb_block1 = molecule._read_and_fix_pdb(path1)
        pdb_block2 = molecule._read_and_fix_pdb(path2)

        compare_blocks(ref_pdb_block, pdb_block1, (76, 78))
        compare_blocks(ref_pdb_block, pdb_block2, (76, 78))

        # Deactivate fixer
        molecule = Molecule(fix_pdb=False)

        ref_path = get_data_file_path('tests/ligSUV_fixed.pdb')
        path1 = get_data_file_path('tests/ligSUV_no_elements1.pdb')
        path2 = get_data_file_path('tests/ligSUV_no_elements2.pdb')

        ref_pdb_block = molecule._read_and_fix_pdb(ref_path)
        pdb_block1 = molecule._read_and_fix_pdb(path1)
        pdb_block2 = molecule._read_and_fix_pdb(path2)

        with pytest.raises(AssertionError):
            compare_blocks(ref_pdb_block, pdb_block1, (76, 78))

        with pytest.raises(AssertionError):
            compare_blocks(ref_pdb_block, pdb_block2, (76, 78))

    def test_pdb_fixer_logger_messages(self):
        """It checks the logger messages of the PDB fixer."""

        from peleffy.utils import Logger
        import io

        molecule = Molecule(fix_pdb=True)

        # Check logger messages
        path3 = get_data_file_path('tests/ligSUV_no_elements3.pdb')

        import logging

        # Force a hard reset of logging library and the logger it manages
        from importlib import reload
        logging.shutdown()
        reload(logging)

        # Initiate logger
        log = Logger()

        # Try the default level (INFO)
        # Catch logger messages to string buffer
        with io.StringIO() as buf:
            # Add custom handler to logger
            log_handler = logging.StreamHandler(buf)
            log._logger.handlers = list()
            log._logger.addHandler(log_handler)

            _ = molecule._read_and_fix_pdb(path3)

            # Get string from buffer
            output = buf.getvalue()

            assert output == "Warning: input PDB has no information " \
                + "about atom elements and they were inferred from " \
                + "atom names. " \
                + "Please, verify that the resulting elements are " \
                + "correct\n" \
                + "Error: PDB could not be fixed\n"

    def test_molecule_display(self):
        """
        It checks the visual representation of the molecule in a
        Jupyter notebook.
        """

        from IPython.display import display

        molecule = Molecule(smiles='c1ccccc1')

        # This should not raise any Exception
        display(molecule)



"""
This module handles all classes and functions related with rotamers.
"""

from collections import defaultdict
import networkx as nx
from copy import deepcopy

from peleffy.utils.toolkits import RDKitToolkitWrapper


class Rotamer(object):
    """
    It represents the conformational space of a rotatable bond, discretized
    with a certain resolution.
    """

    def __init__(self, index1, index2, resolution=30):
        """
        It initiates an Rotamer object.

        Parameters
        ----------
        atom1 : int
            The index of the first atom involved in the rotamer
        atom2 : int
            The index of the second atom involved in the rotamer
        resolution : float
            The resolution to discretize the rotamer's conformational space
        """
        self._index1 = index1
        self._index2 = index2
        self._resolution = resolution

    def __eq__(self, other):
        """Define equality operator"""
        return (self.index1 == other.index1
                and self.index2 == other.index2) \
            or (self.index1 == other.index2
                and self.index2 == other.index1)

    @property
    def index1(self):
        """
        Rotamer's atom1 index.

        Returns
        -------
        index1 : int
            The index of the first atom involved in this Rotamer object
        """
        return self._index1

    @property
    def index2(self):
        """
        Rotamer's atom2 index.

        Returns
        -------
        index2 : int
            The index of the second atom involved in this Rotamer object
        """
        return self._index2

    @property
    def resolution(self):
        """
        Rotamer's resolution.

        Returns
        -------
        resolution : float
            The resolution of this Rotamer object
        """
        return self._resolution


class RotamerLibrary(object):
    """
    It represents a set of rotamers found in the same molecule.
    """

    def __init__(self, molecule):
        """
        It initiates a RotamerLibrary object.

        Parameters
        ----------
        molecule : An peleffy.topology.Molecule
            The Molecule object whose rotamer library will be generated

        Load a molecule and create its rotamer library template

        >>> from peleffy.topology import Molecule
        >>> from peleffy.topology import RotamerLibrary

        >>> molecule = Molecule(smiles='CCCC', name='butane', tag='BUT')

        >>> rotamer_library = RotamerLibrary(mol)
        >>> rotamer_library.to_file('butz')

        Load a molecule and create its rotamer library template with
        a core constraint

        >>> from peleffy.topology import Molecule
        >>> from peleffy.topology import RotamerLibrary

        >>> molecule = Molecule(smiles='CCCC', name='butane', tag='BUT',
                                exclude_terminal_rotamers=False,
                                core_constraints=[0, ])

        >>> rotamer_library = RotamerLibrary(mol)
        >>> rotamer_library.to_file('butz')

        """
        self._molecule = molecule

    def to_file(self, path):
        """
        It writes this RotamerLibrary to a file.

        Parameters
        ----------
        path : str
            Path to save the RotamerLibrary to
        """
        # PELE needs underscores instead of whitespaces
        pdb_atom_names = [name.replace(' ', '_',)
                          for name in self.molecule.get_pdb_atom_names()]

        with open(path, 'w') as file:
            file.write('rot assign res {} &\n'.format(self.molecule.tag))
            for i, rotamer_branches in enumerate(self.molecule.rotamers):
                if i > 0:
                    file.write('     newgrp &\n')
                for rotamer in rotamer_branches:
                    atom_name1 = pdb_atom_names[rotamer.index1]
                    atom_name2 = pdb_atom_names[rotamer.index2]
                    file.write('   sidelib FREE{} {} {} &\n'.format(
                        rotamer.resolution, atom_name1, atom_name2))

    @property
    def molecule(self):
        """
        The peleffy's Molecule.

        Returns
        -------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object
        """
        return self._molecule

    def _ipython_display_(self):
        """
        It displays a 2D molecular representation with bonds highlighted
        according to this rotamer library object.

        Returns
        -------
        representation_2D : a IPython display object
            It is displayable RDKit molecule with an embeded 2D
            representation
        """
        COLORS = [(82 / 255, 215 / 255, 255 / 255), (255 / 255, 154 / 255, 71 / 255),
                  (161 / 255, 255 / 255, 102 / 255), (255 / 255, 173 / 255, 209 / 255),
                  (154 / 255, 92 / 255, 255 / 255), (66 / 255, 255 / 255, 167 / 255),
                  (251 / 255, 255 / 255, 17 / 255)]

        from IPython.display import display

        # Get 2D molecular representation
        rdkit_toolkit = RDKitToolkitWrapper()
        representation = rdkit_toolkit.get_2D_representation(self.molecule)

        # Get rotamer branches from molecule
        rotamer_branches = self.molecule.rotamers

        bond_indexes = list()
        bond_color_dict = dict()
        for bond in representation.GetBonds():
            rotamer = Rotamer(bond.GetBeginAtom().GetIdx(),
                              bond.GetEndAtom().GetIdx())

            for color_index, group in enumerate(rotamer_branches):
                if rotamer in group:
                    bond_indexes.append(bond.GetIdx())
                    try:
                        bond_color_dict[bond.GetIdx()] = COLORS[color_index]
                    except IndexError:
                        bond_color_dict[bond.GetIdx()] = (99 / 255,
                                                          122 / 255,
                                                          126 / 255)
                    break

        atom_indexes = list()
        radii_dict = dict()
        atom_color_dict = dict()

        for atom in representation.GetAtoms():
            atom_index = atom.GetIdx()
            if atom_index in self.molecule._graph.core_nodes:
                atom_indexes.append(atom.GetIdx())
                radii_dict[atom.GetIdx()] = 0.6
                atom_color_dict[atom.GetIdx()] = (255 / 255, 243 / 255, 133 / 255)

        # Get 2D molecular representation
        rdkit_toolkit = RDKitToolkitWrapper()
        representation = rdkit_toolkit.get_2D_representation(self.molecule)

        # Get its image
        image = rdkit_toolkit.draw_molecule(representation,
                                            atom_indexes=atom_indexes,
                                            radii_dict=radii_dict,
                                            atom_color_dict=atom_color_dict,
                                            bond_indexes=bond_indexes,
                                            bond_color_dict=bond_color_dict)

        return display(image)


class MolecularGraph(nx.Graph):
    """
    It represents the structure of a Molecule as a networkx.Graph.
    """

    def __init__(self, molecule):
        """
        It initializes a MolecularGraph object.

        Parameters
        ----------
        molecule : An peleffy.topology.Molecule
            A Molecule object to be written as an Impact file
        """
        super().__init__(self)
        self._molecule = molecule
        self._compute_rotamer_graph()
        self._build_core_nodes()

    def _compute_rotamer_graph(self):
        """
        It initializes the netwrokx.Graph with a Molecule object.
        """
        rdkit_toolkit = RDKitToolkitWrapper()
        rot_bonds_atom_ids = \
            rdkit_toolkit.get_atom_ids_with_rotatable_bonds(self.molecule)

        rdkit_molecule = self.molecule.rdkit_molecule

        atom_names = rdkit_toolkit.get_atom_names(self.molecule)

        assert len(atom_names) == len(rdkit_molecule.GetAtoms()), \
            'The length of atom names must match the length of ' \
            + 'molecule\'s atoms'

        for atom, name in zip(rdkit_molecule.GetAtoms(), atom_names):
            self.add_node(atom.GetIdx(), pdb_name=name,
                          nrot_neighbors=list())

        for bond in rdkit_molecule.GetBonds():
            atom1 = bond.GetBeginAtomIdx()
            atom2 = bond.GetEndAtomIdx()

            if (frozenset([atom1, atom2]) in rot_bonds_atom_ids):
                rotatable = True
            else:
                rotatable = False
                self.nodes[atom1]['nrot_neighbors'].append(atom2)
                self.nodes[atom2]['nrot_neighbors'].append(atom1)

            self.add_edge(bond.GetBeginAtomIdx(),
                          bond.GetEndAtomIdx(),
                          weight=int(rotatable))

        for i, j in rot_bonds_atom_ids:
            self[i][j]['weight'] = 1
            self.nodes[i]['rotatable'] = True
            self.nodes[j]['rotatable'] = True

    def _build_core_nodes(self):
        """
        It builds the list of core nodes
        """

        def get_all_nrot_neighbors(self, atom_id, visited_neighbors):
            """
            A recursive function that hierarchically visits all atom neighbors
            in the graph.

            Parameters
            ----------
            atom_id : int
                Is is both the id of the graph's node and index of the
                corresponding atom
            visited_neighbors : set[int]
                The ids of the nodes that have already been visited

            Returns
            -------
            visited_neighbors : set[int]
                The updated set that contains the ids of the nodes that have
                already been visited
            """
            if atom_id in visited_neighbors:
                return visited_neighbors
            visited_neighbors.add(atom_id)
            nrot_neighbors = self.nodes[atom_id]['nrot_neighbors']
            for nrot_neighbor in nrot_neighbors:
                visited_neighbors = get_all_nrot_neighbors(
                    self, nrot_neighbor, visited_neighbors)
            return visited_neighbors

        from networkx.algorithms.shortest_paths.generic import \
            shortest_path_length
        from networkx.algorithms.distance_measures import eccentricity

        # Calculate graph distances according to weight values
        weighted_distances = dict(shortest_path_length(self, weight="weight"))

        # Calculate eccentricites using weighted distances
        eccentricities = eccentricity(self, sp=weighted_distances)

        # Group nodes by eccentricity
        nodes_by_eccentricities = defaultdict(list)
        for node, ecc in eccentricities.items():
            nodes_by_eccentricities[ecc].append(node)

        # Core atoms must have the minimum eccentricity
        _, centered_nodes = sorted(nodes_by_eccentricities.items())[0]

        # Construct nrot groups with centered nodes
        # already_visited = set()
        centered_node_groups = list()
        for node in centered_nodes:
            # if node in already_visited:
            #    continue
            centered_node_groups.append(get_all_nrot_neighbors(self, node,
                                                               set()))

        # In case of more than one group, core will be the largest
        core_nodes = sorted(centered_node_groups, key=len, reverse=True)[0]

        # To do: think on what to do with the code below
        """
        # Core can hold a maximum of one rotatable bond <- Not true!
        # Get all core's neighbors
        neighbor_candidates = set()
        for node in core_nodes:
            neighbors = self.neighbors(node)
            for neighbor in neighbors:
                if neighbor not in core_nodes:
                    neighbor_candidates.add(neighbor)

        # If any core's neighbor, get the deepest one and include it to
        # the core
        if len(neighbor_candidates) > 0:
            branch_graph = deepcopy(self)

            for node in core_nodes:
                branch_graph.remove_node(node)

            branch_groups = list(nx.connected_components(branch_graph))

            rot_bonds_per_group = self._get_rot_bonds_per_group(branch_groups)

            best_group = sorted(rot_bonds_per_group, key=len,
                                reverse=True)[0]

            for neighbor in neighbor_candidates:
                if any([neighbor in rot_bond for rot_bond in best_group]):
                    deepest_neighbor = neighbor
                    break
            else:
                raise Exception('Unconsistent graph')

            deepest_neighbors = get_all_nrot_neighbors(self, deepest_neighbor,
                                                       set())

            for neighbor in deepest_neighbors:
                core_nodes.add(neighbor)
        """

        self._core_nodes = core_nodes

    def get_parents(self, parent):
        """
        It sets the parent of each atom according to the molecular graph.

        Parameters
        ----------
        parent : int
            The index of the node to use as the absolute parent

        Returns
        -------
        parents : dict[int, int]
            A dictionary containing the index of the parent of each
            atom according to the molecular graph, keyed by the index
            of each child
        """

        def recursive_child_visitor(parent, parents,
                                    already_visited=set()):
            """
            A recursive function that hierarchically visits all the childs of
            each atom.

            Parameters
            ----------
            parent : int
                The index of the atom whose childs will be visited
            parents : dict[int, int]
                A dictionary containing the index of the parent of each
                atom according to the molecular graph, keyed by the index
            visited_neighbors : set[int]
                The updated set that contains the indexes of the atoms
                that have already been visited

            Returns
            -------
            parents : dict[int, int]
                A dictionary containing the index of the parent of each
                atom according to the molecular graph, keyed by the index
                of each child
            visited_neighbors : set[int]
                The updated set that contains the indexes of the atoms
                that have already been visited
            """
            if parent in already_visited:
                return already_visited

            already_visited.add(parent)

            childs = self.neighbors(parent)

            for child in childs:
                if child in already_visited:
                    continue
                parents[child] = parent
                parents, already_visited = recursive_child_visitor(
                    child, parents, already_visited)

            return parents, already_visited

        # Initialize the parents dictionary
        parents = {parent: None}

        parents, already_visited = recursive_child_visitor(parent, parents)

        # Assert absolut parent is the only with a None parent value
        if parents[parent] is not None or \
                sum([int(parents[i] is not None) for i in self.nodes]) \
                != len(self.nodes) - 1:

            from peleffy.utils import Logger
            logger = Logger()
            logger.error('Error: found descendant without parent')

        return parents

    def _get_rot_bonds_per_group(self, branch_groups):
        """
        It constructs the rotatable bonds of each branch group.

        Parameters
        ----------
        branch_groups : list[list[int]]
            The node ids of each branch

        Returns
        -------
        rot_bonds_per_group : list[tuple[int, int]]
            The atom ids of all the graph's edges that belong to a rotatable
            bond
        """
        rot_bonds_per_group = list()
        for group in branch_groups:
            rot_bonds = list()
            visited_bonds = set()
            for node in group:
                bonds = self.edges(node)
                for bond in bonds:
                    if bond in visited_bonds:
                        continue
                    if self[bond[0]][bond[1]]['weight'] == 1:
                        rot_bonds.append(bond)
                    visited_bonds.add(bond)
                    visited_bonds.add((bond[1], bond[0]))
            rot_bonds_per_group.append(rot_bonds)

        return rot_bonds_per_group

    def _get_core_atom_per_group(self, rot_bonds_per_group):
        """
        It obtains the core atom for each group.

        Parameters
        ----------
        rot_bonds_per_group : list[tuple[int, int]]
            The atom ids of all the graph's edges that belong to a rotatable
            bond

        Returns
        -------
        core_atom_per_group : list[int]
            The atom id of the atom that belongs to the core for each branch
        """
        core_atom_per_group = list()
        for rot_bonds in rot_bonds_per_group:
            for (a1, a2) in rot_bonds:
                if a1 in self.core_nodes:
                    core_atom_per_group.append(a1)
                    break
                elif a2 in self.core_nodes:
                    core_atom_per_group.append(a2)
                    break
            else:
                core_atom_per_group.append(None)

        return core_atom_per_group

    def _get_sorted_bonds_per_group(self, core_atom_per_group,
                                    rot_bonds_per_group, distances):
        """
        It sorts in increasing order the rotamers of each group according
        to their distance with respect to the corresponding core atom.

        Parameters
        ----------
        core_atom_per_group : list[int]
            The atom id of the atom that belongs to the core for each branch
        rot_bonds_per_group : list[tuple[int, int]]
            The atom ids of all the graph's edges that belong to a rotatable
            bond
        distances : dict[int, dict[int, int]]
            The distance between each pair of nodes (or atoms)

        Returns
        -------
        sorted_rot_bonds_per_group : list[list]
            The rotatable bonds per group, sorted in increasing order by
            their distance with respect to the corresponding core atom
        """
        sorted_rot_bonds_per_group = list()
        for core_atom, rot_bonds in zip(core_atom_per_group,
                                        rot_bonds_per_group):
            sorting_dict = dict()
            for bond in rot_bonds:
                min_d = min([distances[core_atom][bond[0]],
                             distances[core_atom][bond[1]]])
                sorting_dict[bond] = min_d

            sorted_rot_bonds_per_group.append(
                [i[0] for i in
                 sorted(sorting_dict.items(), key=lambda item: item[1])])

        return sorted_rot_bonds_per_group

    def _ignore_terminal_rotatable_bonds(self, sorted_rot_bonds_per_group,
                                         distances):
        """
        It ignores a certain number of terminal rotatable bonds of each
        group.

        Parameters
        ----------
        sorted_rot_bonds_per_group : list[list]
            The rotatable bonds per group, sorted in increasing order by
            their distance with respect to the corresponding core atom
        distances : dict[int, dict[int, int]]
            The distance between each pair of nodes (or atoms)

        Returns
        -------
        filtered_rot_bonds_per_group : list[list]
            The filtered rotatable bonds per group that are obtained
        """
        filtered_rot_bonds_per_group = list()

        # To determine the outter atom in a rotamer, we only need to
        # calculate their distance to any core atom
        core_atom = list(self.core_nodes)[0]

        for rot_bonds in sorted_rot_bonds_per_group:
            rotamer_to_evaluate = rot_bonds[-1]

            node1, node2 = rotamer_to_evaluate

            distance1 = distances[core_atom][node1]
            distance2 = distances[core_atom][node2]

            if distance1 < distance2:
                inner_node = node1
                outter_node = node2
            else:
                inner_node = node2
                outter_node = node1

            # The condition for the current rotamer to be ignored is that the
            # outter node is only attached to terminal nodes (with degree 1)
            ignore_this_rotamer = True
            for neighbor in self.neighbors(outter_node):
                if neighbor is inner_node:
                    continue

                if self.degree(neighbor) > 1:
                    ignore_this_rotamer = False
                    break

            if ignore_this_rotamer:
                filtered_rot_bonds_per_group.append(rot_bonds[:-1])
            else:
                filtered_rot_bonds_per_group.append(rot_bonds)

        return filtered_rot_bonds_per_group

    def get_rotamers(self):
        """
        It builds the RotamerLibrary object.

        Returns
        -------
        rotamers : list[list]
            The list of rotamers grouped by the branch they belong to
        """
        resolution = self.molecule.rotamer_resolution

        assert len(self.core_nodes) > 0, 'No core nodes were found'

        branch_graph = deepcopy(self)

        for node in self.core_nodes:
            branch_graph.remove_node(node)

        branch_groups = list(nx.connected_components(branch_graph))

        rot_bonds_per_group = self._get_rot_bonds_per_group(branch_groups)

        core_atom_per_group = self._get_core_atom_per_group(
            rot_bonds_per_group)

        distances = dict(nx.shortest_path_length(self))

        sorted_rot_bonds_per_group = self._get_sorted_bonds_per_group(
            core_atom_per_group, rot_bonds_per_group, distances)

        """
        if not include_terminal_rotamers:
            sorted_rot_bonds_per_group = \
                self._ignore_terminal_rotatable_bonds(
                    sorted_rot_bonds_per_group, distances)
        """

        rotamers = list()

        # TODO extend core by including one rotatable bond from the largest
        # branch to increase the performance of the algorithm.
        for group_id, rot_bonds in enumerate(sorted_rot_bonds_per_group):
            branch_rotamers = list()
            for (atom1_index, atom2_index) in rot_bonds:
                rotamer = Rotamer(atom1_index, atom2_index, resolution)
                branch_rotamers.append(rotamer)

            if len(branch_rotamers) > 0:
                rotamers.append(branch_rotamers)

        return rotamers

    @property
    def molecule(self):
        """
        The peleffy's Molecule.

        Returns
        -------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object
        """
        return self._molecule

    @property
    def core_nodes(self):
        """
        The list of core nodes.

        Returns
        -------
        core_nodes : list[int]
            The nodes in the core
        """
        return self._core_nodes


class MolecularGraphWithConstrainedCore(MolecularGraph):
    """
    It represents the structure of a Molecule as a networkx.Graph.
    """

    def __init__(self, molecule, atom_constraints):
        """
        It initializes a MolecularGraph object.

        Parameters
        ----------
        molecule : An peleffy.topology.Molecule
            A Molecule object to be written as an Impact file
        atom_constraint : list[int or str]
            It defines the list of atoms to constrain in the core, thus,
            the core will be forced to contain them. Atoms can be specified
            through integers that match the atom index or strings that
            match with the atom PDB name

        Raises
        ------
        ValueError
            If the supplied array of atom constraints is empty
            If the PDB atom name in atom_constraint does not match with
            any atom in the molecule
        TypeError
            If the atom_constraint is of invalid type
        """
        if len(atom_constraints) == 0:
            raise ValueError('Supplied empty array of atom constraints')

        self._constraint_indices = list()

        for atom_constraint in atom_constraints:
            if isinstance(atom_constraint, int):
                self._constraint_indices.append(atom_constraint)
            elif isinstance(atom_constraint, str):
                atom_names = molecule.get_pdb_atom_names()
                for index, name in enumerate(atom_names):
                    if name == atom_constraint:
                        self._constraint_indices.append(index)
                        break
                else:
                    raise ValueError('Supplied PDB atom name '
                                     + '\'{}\''.format(atom_constraint)
                                     + 'is missing in molecule')
            else:
                raise TypeError('Invalid type for the atom_constraint')

        super().__init__(molecule)

        self._safety_check()

    def _build_core_nodes(self):
        """
        It builds the list of core nodes
        """

        from networkx.algorithms.shortest_paths.generic import \
            shortest_path_length

        # Force core to contain constrained atoms
        core_nodes = [index for index in self.constraint_indices]

        # Calculate graph distances according to weight values
        weighted_distances = dict(shortest_path_length(self, weight="weight"))

        # Add also all atoms at 0 distance with respect to constrained
        # atom into the core
        for node in self.nodes:
            for constraint_index in self.constraint_indices:
                d = weighted_distances[constraint_index][node]
                if d == 0 and node not in core_nodes:
                    core_nodes.append(node)

        self._core_nodes = core_nodes

    def _safety_check(self):
        """Perform a safety check on the atom constraints."""
        if len(self.constraint_indices) < 2:
            return

        safe_indices = set()
        for i, cidx1 in enumerate(self.constraint_indices):
            if cidx1 in safe_indices:
                continue
            for cidx2 in self.constraint_indices:
                if cidx2 in self.neighbors(cidx1):
                    safe_indices.add(cidx1)
                    safe_indices.add(cidx2)
                    break
            else:
                raise ValueError('All atoms in atom constraints must be '
                                 + 'adjacent and atom '
                                 + self.constraint_names[i].strip()
                                 + ' is not')

    @property
    def constraint_indices(self):
        """
        The indices of atoms to constraint to the core.

        Returns
        -------
        constraint_indices : list[int]
            List of atom indices
        """
        return self._constraint_indices

    @property
    def constraint_names(self):
        """
        The names of atoms to constraint to the core.

        Returns
        -------
        constraint_names : list[str]
            List of atom names
        """
        atom_names = self.molecule.get_pdb_atom_names()

        constraint_names = list()
        for index in self.constraint_indices:
            constraint_names.append(atom_names[index])

        return constraint_names



"""
This module contains tests that check that the BCEConformations module.
"""

import numpy as np
import pytest


def build_mock_BCEConformations(pdb_file, ffld_file):
    from peleffy.topology import Molecule, BCEConformations
    from peleffy.forcefield import ForceFieldSelector
    from peleffy.topology import Topology
    from .utils import parameterize_opls2005

    molecule = Molecule(pdb_file)
    ff_selector = ForceFieldSelector()
    forcefield = ff_selector.get_by_name("opls2005")
    parameters = parameterize_opls2005(forcefield, molecule, ffld_file, charge_method="opls2005")
    topology = Topology(molecule, parameters)
    return BCEConformations(topology, "")


class TestBCEConformations(object):
    """BCEConformations test."""

    from peleffy.topology import BCEConformations

    def test_calculate_cluster_offsets(self):
        from peleffy.utils import get_data_file_path

        pdb_path = get_data_file_path('ligands/ethylene.pdb')
        ffld_path = get_data_file_path("tests/ETL_ffld_output.txt")
        golden_offsets = [["_C1_", 0.0, 0.0, 0.0],
                          ["_C2_", -1.305,  0.181, -0.014],
                          ["_H1_", 0.72, 0.822, 0.042],
                          ["_H2_", 0.448, -0.988, -0.029],
                          ["_H3_", -1.73, 1.171, 0.015],
                          ["_H4_", -1.936, -0.687, -0.056]]
        bce_obj = build_mock_BCEConformations(pdb_path, ffld_path)
        bce_obj.calculate_cluster_offsets(pdb_path)
        for dih1, dih2 in zip(bce_obj.conformations_library[pdb_path], golden_offsets):
            assert dih1[0] == dih2[0]
            np.testing.assert_almost_equal(dih1[1:], dih2[1:], decimal=3)

    def test_write_dihedral_library(self):
        import os
        import tempfile
        from peleffy.utils import get_data_file_path, temporary_cd

        pdb_path = get_data_file_path('ligands/ethylene.pdb')
        ffld_path = get_data_file_path("tests/ETL_ffld_output.txt")
        bce_obj = build_mock_BCEConformations(pdb_path, ffld_path)
        bce_obj.calculate_cluster_offsets(pdb_path)
        calculated_lines = []
        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                bce_obj.save(os.path.join(tmpdir, "ETH.conformation"))
                with open(os.path.join(tmpdir, "ETH.conformation")) as f:
                    calculated_lines = f.readlines()
        calculated_lines = calculated_lines[3:]
        golden_conformation_library_path = get_data_file_path('parameters/ETH.conformation')
        with open(golden_conformation_library_path) as f:
            golden_lines = f.readlines()[3:]

        assert golden_lines == calculated_lines



"""
This module contains the tests to check the peleffy's rotamer library
builder.
"""

import pytest

from peleffy.utils import get_data_file_path
from peleffy.topology import Molecule


class TestMolecularGraph(object):
    """
    It wraps all tests that check the MolecularGraph class.
    """

    def test_rotamer_library_builder(self):
        """
        It tests the rotamer library builder.
        """
        LIGAND_PATH = 'ligands/oleic_acid.pdb'

        ligand_path = get_data_file_path(LIGAND_PATH)
        molecule = Molecule(ligand_path, exclude_terminal_rotamers=False)

        # rotamer_library = RotamerLibrary(molecule)

        rotamers_per_branch = molecule.rotamers

        assert len(rotamers_per_branch) == 2, "Found an invalid number " + \
            "of branches: {}".format(len(rotamers_per_branch))

        atom_list_1 = list()
        atom_list_2 = list()

        rotamers = rotamers_per_branch[0]
        for rotamer in rotamers:
            atom_list_1.append(set([rotamer.index1, rotamer.index2]))

        rotamers = rotamers_per_branch[1]
        for rotamer in rotamers:
            atom_list_2.append(set([rotamer.index1, rotamer.index2]))

        EXPECTED_INDICES_1 = [set([9, 10]), set([8, 9]), set([7, 8]),
                              set([6, 7]), set([5, 6]), set([2, 5]),
                              set([0, 2]), set([0, 1])]

        EXPECTED_INDICES_2 = [set([12, 11]), set([12, 13]), set([13, 14]),
                              set([14, 15]), set([15, 16]), set([16, 17]),
                              set([17, 18]), set([18, 19])]

        where_1 = list()
        for atom_pair in atom_list_1:
            if atom_pair in EXPECTED_INDICES_1:
                where_1.append(1)
            elif atom_pair in EXPECTED_INDICES_2:
                where_1.append(2)
            else:
                where_1.append(0)

        where_2 = list()
        for atom_pair in atom_list_2:
            if atom_pair in EXPECTED_INDICES_1:
                where_2.append(1)
            elif atom_pair in EXPECTED_INDICES_2:
                where_2.append(2)
            else:
                where_2.append(0)

        assert (all(i == 1 for i in where_1)
                and all(i == 2 for i in where_2)) or \
            (all(i == 2 for i in where_1)
             and all(i == 1 for i in where_2)), "Invalid rotamer library " + \
            "{}, {}".format(where_1, where_2)

        assert (all(i == 1 for i in where_1)
                and all(i == 2 for i in where_2)
                and len(where_1) == len(EXPECTED_INDICES_1)
                and len(where_2) == len(EXPECTED_INDICES_2)) or \
               (all(i == 2 for i in where_1)
                and all(i == 1 for i in where_2)
                and len(where_1) == len(EXPECTED_INDICES_2)
                and len(where_2) == len(EXPECTED_INDICES_1)), "Unexpected " + \
            "number of rotamers"

    def test_terminal_rotamer_filtering(self):
        """
        It tests the rotamer library builder when the terminal rotatable bonds
        are ignored.
        """
        LIGAND_PATH = 'ligands/oleic_acid.pdb'

        ligand_path = get_data_file_path(LIGAND_PATH)
        molecule = Molecule(ligand_path, exclude_terminal_rotamers=True)

        rotamers_per_branch = molecule.rotamers

        assert len(rotamers_per_branch) == 2, "Found an invalid number " + \
            "of branches: {}".format(len(rotamers_per_branch))

        atom_list_1 = list()
        atom_list_2 = list()
        rotamers = rotamers_per_branch[0]
        for rotamer in rotamers:
            atom_list_1.append(set([rotamer.index1, rotamer.index2]))

        rotamers = rotamers_per_branch[1]
        for rotamer in rotamers:
            atom_list_2.append(set([rotamer.index1, rotamer.index2]))

        EXPECTED_INDICES_1 = [set([9, 10]), set([8, 9]), set([7, 8]),
                              set([6, 7]), set([5, 6]), set([2, 5]),
                              set([0, 2]), set([0, 1])]

        EXPECTED_INDICES_2 = [set([12, 11]), set([12, 13]), set([13, 14]),
                              set([14, 15]), set([15, 16]), set([16, 17]),
                              set([17, 18])]

        where_1 = list()
        for atom_pair in atom_list_1:
            if atom_pair in EXPECTED_INDICES_1:
                where_1.append(1)
            elif atom_pair in EXPECTED_INDICES_2:
                where_1.append(2)
            else:
                where_1.append(0)

        where_2 = list()
        for atom_pair in atom_list_2:
            if atom_pair in EXPECTED_INDICES_1:
                where_2.append(1)
            elif atom_pair in EXPECTED_INDICES_2:
                where_2.append(2)
            else:
                where_2.append(0)

        assert (all(i == 1 for i in where_1)
                and all(i == 2 for i in where_2)) or \
            (all(i == 2 for i in where_1)
             and all(i == 1 for i in where_2)), "Invalid rotamer library " + \
            "{}, {}".format(where_1, where_2)

        assert (all(i == 1 for i in where_1)
                and all(i == 2 for i in where_2)
                and len(where_1) == len(EXPECTED_INDICES_1)
                and len(where_2) == len(EXPECTED_INDICES_2)) or \
               (all(i == 2 for i in where_1)
                and all(i == 1 for i in where_2)
                and len(where_1) == len(EXPECTED_INDICES_2)
                and len(where_2) == len(EXPECTED_INDICES_1)), "Unexpected " + \
            "number of rotamers"

    def test_rotamer_core_constraint(self):
        """
        It tests the rotamer library builder when constraining its core
        to contain a specific atom.
        """

        LIGAND_PATH = 'ligands/oleic_acid.pdb'
        ligand_path = get_data_file_path(LIGAND_PATH)

        # Test atom index constraint
        molecule = Molecule(ligand_path, core_constraints=[19, ],
                            exclude_terminal_rotamers=False)

        rotamers_per_branch = molecule.rotamers

        assert len(rotamers_per_branch) == 1, "Found an invalid number " + \
            "of branches: {}".format(len(rotamers_per_branch))

        atom_list = list()
        for rotamer in rotamers_per_branch[0]:
            atom_list.append(set([rotamer.index1, rotamer.index2]))

        EXPECTED_INDICES = [set([18, 19]), set([17, 18]), set([16, 17]),
                            set([15, 16]), set([14, 15]), set([13, 14]),
                            set([12, 13]), set([11, 12]), set([9, 10]),
                            set([8, 9]), set([7, 8]), set([6, 7]),
                            set([5, 6]), set([2, 5]), set([0, 2]),
                            set([0, 1])]

        assert len(atom_list) == len(EXPECTED_INDICES), "Unexpected " + \
            "number of rotamers"

        assert all(atom_pair in EXPECTED_INDICES for atom_pair in atom_list), \
            "Invalid rotamer library"

        # Test PDB atom name constraint
        molecule = Molecule(ligand_path, core_constraints=[' C18', ],
                            exclude_terminal_rotamers=False)

        rotamers_per_branch = molecule.rotamers

        assert len(rotamers_per_branch) == 1, "Found an invalid number " + \
            "of branches: {}".format(len(rotamers_per_branch))

        atom_list = list()
        for rotamer in rotamers_per_branch[0]:
            atom_list.append(set([rotamer.index1, rotamer.index2]))

        EXPECTED_INDICES = [set([18, 19]), set([17, 18]), set([16, 17]),
                            set([15, 16]), set([14, 15]), set([13, 14]),
                            set([12, 13]), set([11, 12]), set([9, 10]),
                            set([8, 9]), set([7, 8]), set([6, 7]),
                            set([5, 6]), set([2, 5]), set([0, 2]),
                            set([0, 1])]

        assert len(atom_list) == len(EXPECTED_INDICES), "Unexpected " + \
            "number of rotamers"

        assert all(atom_pair in EXPECTED_INDICES for atom_pair in atom_list), \
            "Invalid rotamer library"

        # Test core constraint with terminal exclusion
        molecule = Molecule(ligand_path, core_constraints=[' C18', ],
                            exclude_terminal_rotamers=True)

        rotamers_per_branch = molecule.rotamers

        assert len(rotamers_per_branch) == 1, "Found an invalid number " + \
            "of branches: {}".format(len(rotamers_per_branch))

        atom_list = list()
        for rotamer in rotamers_per_branch[0]:
            atom_list.append(set([rotamer.index1, rotamer.index2]))

        EXPECTED_INDICES = [set([17, 18]), set([16, 17]), set([15, 16]),
                            set([14, 15]), set([13, 14]), set([12, 13]),
                            set([11, 12]), set([9, 10]), set([8, 9]),
                            set([7, 8]), set([6, 7]), set([5, 6]),
                            set([2, 5]), set([0, 2]), set([0, 1])]

        assert len(atom_list) == len(EXPECTED_INDICES), "Unexpected " + \
            "number of rotamers"

        assert all(atom_pair in EXPECTED_INDICES for atom_pair in atom_list), \
            "Invalid rotamer library"

        # Test core constraint with a central core
        molecule = Molecule(ligand_path, core_constraints=[' C9 ', ],
                            exclude_terminal_rotamers=True)

        rotamers_per_branch = molecule.rotamers

        assert len(rotamers_per_branch) == 2, "Found an invalid number " + \
            "of branches: {}".format(len(rotamers_per_branch))

        atom_list_1 = list()
        atom_list_2 = list()
        rotamers = rotamers_per_branch[0]
        for rotamer in rotamers:
            atom_list_1.append(set([rotamer.index1, rotamer.index2]))

        rotamers = rotamers_per_branch[1]
        for rotamer in rotamers:
            atom_list_2.append(set([rotamer.index1, rotamer.index2]))

        EXPECTED_INDICES_1 = [set([9, 10]), set([8, 9]), set([7, 8]),
                              set([6, 7]), set([5, 6]), set([2, 5]),
                              set([0, 2]), set([0, 1])]

        EXPECTED_INDICES_2 = [set([12, 11]), set([12, 13]), set([13, 14]),
                              set([14, 15]), set([15, 16]), set([16, 17]),
                              set([17, 18])]

        where_1 = list()
        for atom_pair in atom_list_1:
            if atom_pair in EXPECTED_INDICES_1:
                where_1.append(1)
            elif atom_pair in EXPECTED_INDICES_2:
                where_1.append(2)
            else:
                where_1.append(0)

        where_2 = list()
        for atom_pair in atom_list_2:
            if atom_pair in EXPECTED_INDICES_1:
                where_2.append(1)
            elif atom_pair in EXPECTED_INDICES_2:
                where_2.append(2)
            else:
                where_2.append(0)

        assert (all(i == 1 for i in where_1)
                and all(i == 2 for i in where_2)) or \
            (all(i == 2 for i in where_1)
             and all(i == 1 for i in where_2)), "Invalid rotamer library " + \
            "{}, {}".format(where_1, where_2)

        assert (all(i == 1 for i in where_1)
                and all(i == 2 for i in where_2)
                and len(where_1) == len(EXPECTED_INDICES_1)
                and len(where_2) == len(EXPECTED_INDICES_2)) or \
               (all(i == 2 for i in where_1)
                and all(i == 1 for i in where_2)
                and len(where_1) == len(EXPECTED_INDICES_2)
                and len(where_2) == len(EXPECTED_INDICES_1)), "Unexpected " + \
            "number of rotamers"

        # Test core constraint with a multiple central core
        molecule = Molecule(ligand_path,
                            core_constraints=[' C8 ', ' C9 ', ' C10'],
                            exclude_terminal_rotamers=True)

        rotamers_per_branch = molecule.rotamers

        assert len(rotamers_per_branch) == 2, "Found an invalid number " + \
            "of branches: {}".format(len(rotamers_per_branch))

        atom_list_1 = list()
        atom_list_2 = list()
        rotamers = rotamers_per_branch[0]
        for rotamer in rotamers:
            atom_list_1.append(set([rotamer.index1, rotamer.index2]))

        rotamers = rotamers_per_branch[1]
        for rotamer in rotamers:
            atom_list_2.append(set([rotamer.index1, rotamer.index2]))

        EXPECTED_INDICES_1 = [set([8, 9]), set([7, 8]), set([6, 7]),
                              set([5, 6]), set([2, 5]), set([0, 2]),
                              set([0, 1])]

        EXPECTED_INDICES_2 = [set([12, 11]), set([12, 13]), set([13, 14]),
                              set([14, 15]), set([15, 16]), set([16, 17]),
                              set([17, 18])]

        where_1 = list()
        for atom_pair in atom_list_1:
            if atom_pair in EXPECTED_INDICES_1:
                where_1.append(1)
            elif atom_pair in EXPECTED_INDICES_2:
                where_1.append(2)
            else:
                where_1.append(0)

        where_2 = list()
        for atom_pair in atom_list_2:
            if atom_pair in EXPECTED_INDICES_1:
                where_2.append(1)
            elif atom_pair in EXPECTED_INDICES_2:
                where_2.append(2)
            else:
                where_2.append(0)

        assert (all(i == 1 for i in where_1)
                and all(i == 2 for i in where_2)) or \
            (all(i == 2 for i in where_1)
             and all(i == 1 for i in where_2)), "Invalid rotamer library " + \
            "{}, {}".format(where_1, where_2)

        assert (all(i == 1 for i in where_1)
                and all(i == 2 for i in where_2)
                and len(where_1) == len(EXPECTED_INDICES_1)
                and len(where_2) == len(EXPECTED_INDICES_2)) or \
               (all(i == 2 for i in where_1)
                and all(i == 1 for i in where_2)
                and len(where_1) == len(EXPECTED_INDICES_2)
                and len(where_2) == len(EXPECTED_INDICES_1)), "Unexpected " + \
            "number of rotamers"

    def test_rotamer_core_constraint_adjacency(self):
        """
        It tests the adjacency check up that is performed prior building
        the rotamer library builder with core constraints.
        """

        LIGAND_PATH = 'ligands/oleic_acid.pdb'
        ligand_path = get_data_file_path(LIGAND_PATH)

        # Test adjacent core constraint selection
        _ = Molecule(ligand_path,
                     core_constraints=[' C8 ', ' C9 ', ' C10'])

        # Test non adjacent core constraint selection
        with pytest.raises(ValueError) as e:
            _ = Molecule(ligand_path,
                         core_constraints=[' C1 ', ' C9 ', ' C10'])

        assert str(e.value) == 'All atoms in atom constraints must be ' \
            + 'adjacent and atom C1 is not'



"""
This module contains the tests to check peleffy's molecular representations.
"""

import pytest

import tempfile
from peleffy.utils import get_data_file_path, temporary_cd


class TestMain(object):
    """
    It wraps all tests that involve the Molecule class.
    """

    def test_peleffy_default_call(self):
        """
        It checks the default call of peleffy's main function.
        """
        from peleffy.main import run_peleffy

        LIGAND_PATH = 'ligands/benzene.pdb'
        ligand_path = get_data_file_path(LIGAND_PATH)

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                run_peleffy(ligand_path, output=tmpdir)

    def test_peleffy_custom_call(self):
        """
        It checks the custom call of peleffy's main function.
        """
        from peleffy.main import run_peleffy

        LIGAND_PATH = 'ligands/benzene.pdb'
        CUSTOM_FORCEFIELD_NAME = 'openff_unconstrained-1.2.0.offxml'

        ligand_path = get_data_file_path(LIGAND_PATH)

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                run_peleffy(ligand_path,
                            forcefield_name=CUSTOM_FORCEFIELD_NAME,
                            resolution=10,
                            charge_method='gasteiger',
                            output=tmpdir,
                            with_solvent=True,
                            as_datalocal=True)

    def test_peleffy_argparse(self):
        """It checks the command-line argument parser of peleffy."""
        from peleffy.main import parse_args

        # Test positional arguments
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            parsed_args = parse_args([])

        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 2

        # Test defaults
        parsed_args = parse_args(['benzene.pdb'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'am1bcc', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.3.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output is None, \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'benzene.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 30, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path is None, \
            'Unexpected conformations_path settings were parsed'

        # Test custom shorts
        parsed_args = parse_args(['toluene.pdb',
                                  '-f', 'openff_unconstrained-1.0.0.offxml',
                                  '-r', '60',
                                  '-o', 'my_custom_output',
                                  '-c', 'gasteiger'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'gasteiger', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.0.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output == 'my_custom_output', \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'toluene.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 60, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path is None, \
            'Unexpected conformations_path settings were parsed'

        # Test custom longs
        parsed_args = parse_args(['methane.pdb',
                                  '--forcefield',
                                  'openff_unconstrained-1.0.1.offxml',
                                  '--resolution', '120',
                                  '--output', 'my_custom_output2',
                                  '--charge_method', 'gasteiger'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'gasteiger', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.0.1.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output == 'my_custom_output2', \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'methane.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 120, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path is None, \
            'Unexpected conformations_path settings were parsed'

        # Test unexpected charge method
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            parsed_args = parse_args(['toluene.pdb', '-c', 'unexpected'])

        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 2

        # Test as_datalocal argument
        parsed_args = parse_args(['methane.pdb',
                                  '--as_datalocal'])

        assert parsed_args.as_datalocal is True, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'am1bcc', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.3.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output is None, \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'methane.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 30, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path is None, \
            'Unexpected conformations_path settings were parsed'

        # Test charges_from_file argument
        parsed_args = parse_args(['BHP.pdb',
                                  '--charges_from_file', 'BHP.mae'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'am1bcc', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.3.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.charges_from_file == 'BHP.mae', \
            'Unexpected charges_from_file settings were parsed'
        assert parsed_args.output is None, \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'BHP.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 30, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path is None, \
            'Unexpected conformations_path settings were parsed'

        # Test include_terminal_rotamers argument
        parsed_args = parse_args(['methane.pdb',
                                  '--include_terminal_rotamers'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'am1bcc', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.3.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is True, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output is None, \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'methane.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 30, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path is None, \
            'Unexpected conformations_path settings were parsed'

        # Test chain argument
        parsed_args = parse_args(['LYS_BNZ.pdb',
                                  '--chain', 'L'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'am1bcc', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.3.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output is None, \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'LYS_BNZ.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 30, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.chain == 'L',\
            'Unexpected chain settings were parsed'

        # Test silent argument
        parsed_args = parse_args(['methane.pdb',
                                  '-s'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'am1bcc', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.3.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output is None, \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'methane.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 30, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is True, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path is None, \
            'Unexpected conformations_path settings were parsed'

        parse_args(['methane.pdb', '-s']) == parse_args(['methane.pdb',
                                                         '--silent'])

        # Test debug argument
        parsed_args = parse_args(['methane.pdb',
                                  '-d'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'am1bcc', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is True, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.3.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output is None, \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'methane.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 30, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path is None, \
            'Unexpected conformations_path settings were parsed'

        parse_args(['methane.pdb', '-d']) == parse_args(['methane.pdb',
                                                         '--debug'])

        # Test dihedral library arguments
        parsed_args = parse_args(['methane.pdb',
                                  '--conformations_info_path', 'test_path'])

        assert parsed_args.as_datalocal is False, \
            'Unexpected as_datalocal settings were parsed'
        assert parsed_args.charge_method == 'am1bcc', \
            'Unexpected charge_method settings were parsed'
        assert parsed_args.debug is False, \
            'Unexpected debug settings were parsed'
        assert parsed_args.forcefield == 'openff_unconstrained-1.3.0.offxml', \
            'Unexpected forcefield settings were parsed'
        assert parsed_args.include_terminal_rotamers is False, \
            'Unexpected include_terminal_rotamers settings were parsed'
        assert parsed_args.output is None, \
            'Unexpected output settings were parsed'
        assert parsed_args.pdb_file == 'methane.pdb', \
            'Unexpected pdb_file settings were parsed'
        assert parsed_args.resolution == 30, \
            'Unexpected resolution settings were parsed'
        assert parsed_args.silent is False, \
            'Unexpected silent settings were parsed'
        assert parsed_args.with_solvent is False, \
            'Unexpected with_solvent settings were parsed'
        assert parsed_args.conformations_info_path == "test_path", \
            'Unexpected conformations_path settings were parsed'

    def test_peleffy_main(self):
        """It checks the main function of peleffy."""
        from peleffy.main import parse_args, main
        from peleffy.utils import Logger
        import logging

        ligand_path = get_data_file_path('ligands/benzene.pdb')

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                # Test default settings
                args = parse_args([ligand_path])
                main(args)

                logger = Logger()
                for handler in logger._logger.handlers:
                    assert handler.level == logging.INFO

                # Test silent settings
                args = parse_args([ligand_path, '--silent'])
                main(args)

                logger = Logger()
                for handler in logger._logger.handlers:
                    assert handler.level == logging.CRITICAL

                # Test silent settings
                args = parse_args([ligand_path, '--debug'])
                main(args)

                logger = Logger()
                for handler in logger._logger.handlers:
                    assert handler.level == logging.DEBUG



from .topology import Topology
from .elements import Atom, Bond, Angle, Dihedral, Proper, Improper
from .molecule import Molecule
from .zmatrix import ZMatrix
from .rotamer import RotamerLibrary
from .conformer import BCEConformations



"""
This module contains the tests to check all available force fields in
peleffy.
"""


class TestTopology(object):
    """
    It wraps all tests that check the OpenForceField class.
    """

    def test_empty_parameters(self):
        """
        It tests the initialization of a Topology object with an empty
        parameters wrapper.
        """

        from peleffy.topology import Molecule
        from peleffy.forcefield.parameters import BaseParameterWrapper
        from peleffy.topology import Topology

        molecule = Molecule()
        parameters = BaseParameterWrapper()
        Topology(molecule, parameters)

    def test_add_topological_elements(self):
        """
        It tests the addition of topological elements to an empty
        topology.
        """

        from peleffy.topology import Molecule
        from peleffy.forcefield import OpenForceField
        from peleffy.forcefield.parameters import BaseParameterWrapper
        from peleffy.topology import Topology
        from peleffy.utils import get_data_file_path

        # Define molecule1 and its topology
        pdb_path = get_data_file_path('ligands/ethylene.pdb')
        molecule1 = Molecule(pdb_path)
        openff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        parameters1 = openff.parameterize(molecule1)
        topology1 = Topology(molecule1, parameters1)

        # Define empty topology2
        molecule2 = Molecule()
        parameters2 = BaseParameterWrapper()
        topology2 = Topology(molecule2, parameters2)

        # Add parameters to topology2
        for atom in topology1.atoms:
            topology2.add_atom(atom)
        for bond in topology1.bonds:
            topology2.add_bond(bond)
        for angle in topology1.angles:
            topology2.add_angle(angle)
        for proper in topology1.propers:
            topology2.add_proper(proper)
        for improper in topology1.impropers:
            topology2.add_improper(improper)

        # Verify content of both topologies
        assert topology1.atoms == topology2.atoms, \
            'The atoms of boths topologies should match'
        assert topology1.bonds == topology2.bonds, \
            'The bonds of boths topologies should match'
        assert topology1.angles == topology2.angles, \
            'The angles of boths topologies should match'
        assert topology1.propers == topology2.propers, \
            'The propers of boths topologies should match'
        assert topology1.impropers == topology2.impropers, \
            'The impropers of boths topologies should match'

    def test_openff_parameterizer(self):
        """
        It checks the behaviour of the Topology with the OpenFF
        parameters.
        """

        from peleffy.topology import Molecule
        from peleffy.forcefield import OpenForceField
        from peleffy.topology import Topology
        from peleffy.utils import get_data_file_path
        from .utils import check_parameters

        FORCE_FIELD_NAME = 'openff_unconstrained-1.2.1.offxml'

        # Load molecule
        molecule = Molecule(get_data_file_path('ligands/methane.pdb'))
        openff = OpenForceField(FORCE_FIELD_NAME)

        # Obtain force field parameters
        parameters = openff.parameterize(molecule)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Define expected parameters
        expected_nonbonding = [
            [1, 0, 'M', 'OFFT', '_C1_', 0, 3.3996695084235347, 0.1094,
             -0.1088, 0, 1.6998347542117673, 0, 0],
            [2, 1, 'M', 'OFFT', '_H1_', 0, 2.649532787749369, 0.0157,
             0.0267, 0, 1.3247663938746845, 0, 0],
            [3, 1, 'M', 'OFFT', '_H2_', 0, 2.649532787749369, 0.0157,
             0.0267, 0, 1.3247663938746845, 0, 0],
            [4, 1, 'M', 'OFFT', '_H3_', 0, 2.649532787749369, 0.0157,
             0.0267, 0, 1.3247663938746845, 0, 0],
            [5, 1, 'M', 'OFFT', '_H4_', 0, 2.649532787749369, 0.0157,
             0.0267, 0, 1.3247663938746845, 0, 0]]

        expected_bonds = [
            [1, 2, 376.8940758588, 1.094223427522],
            [1, 3, 376.8940758588, 1.094223427522],
            [1, 4, 376.8940758588, 1.094223427522],
            [1, 5, 376.8940758588, 1.094223427522]]

        expected_angles = [
            [2, 1, 3, 33.78875634641, 110.2468561538],
            [2, 1, 4, 33.78875634641, 110.2468561538],
            [2, 1, 5, 33.78875634641, 110.2468561538],
            [3, 1, 4, 33.78875634641, 110.2468561538],
            [3, 1, 5, 33.78875634641, 110.2468561538],
            [4, 1, 5, 33.78875634641, 110.2468561538]]

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_nonbonding,
                         expected_bonds=expected_bonds,
                         expected_angles=expected_angles)

        # Load molecule
        molecule = Molecule(get_data_file_path('ligands/ethylene.pdb'))
        openff = OpenForceField(FORCE_FIELD_NAME)

        # Obtain force field parameters
        parameters = openff.parameterize(molecule)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Define expected parameters
        expected_propers = [
            [3, 1, 2, 5, 5.376019778605, -1, 2, 0.0],
            [3, 1, 2, 6, 5.376019778605, -1, 2, 0.0],
            [4, 1, 2, 5, 5.376019778605, -1, 2, 0.0],
            [4, 1, 2, 6, 5.376019778605, -1, 2, 0.0]]

        expected_impropers = [
            [1, 2, 5, 6, 1.1, -1, 2],
            [2, 1, 3, 4, 1.1, -1, 2]]

        # Check it up
        check_parameters(topology,
                         expected_propers=expected_propers,
                         expected_impropers=expected_impropers)

    def test_opls2005_parameterizer(self):
        """
        It checks the behaviour of the Topology with the OPLS2005
        parameters.
        """

        from peleffy.topology import Molecule
        from peleffy.topology import Topology
        from peleffy.forcefield import OPLS2005ForceField
        from peleffy.utils import get_data_file_path
        from .utils import check_parameters, parameterize_opls2005

        # Load molecule
        opls2005 = OPLS2005ForceField()
        molecule = Molecule(get_data_file_path('ligands/methane.pdb'))
        ffld_file = get_data_file_path('tests/MET_ffld_output.txt')

        # Workaround to parameterize with OPLS2005 without the Schrodinger
        # dependency
        parameters = parameterize_opls2005(opls2005, molecule, ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Define expected parameters
        expected_nonbonding = [
            [1, 0, 'M', 'CT', '_C1_', 0, 3.5, 0.066, -0.24, 1.975, 1.75,
             0.005, -0.74168571],
            [2, 1, 'M', 'HC', '_H1_', 0, 2.5, 0.03, 0.06, 1.425, 1.25,
             0.00859824, 0.268726247],
            [3, 1, 'M', 'HC', '_H2_', 0, 2.5, 0.03, 0.06, 1.425, 1.25,
             0.00859824, 0.268726247],
            [4, 1, 'M', 'HC', '_H3_', 0, 2.5, 0.03, 0.06, 1.425, 1.25,
             0.00859824, 0.268726247],
            [5, 1, 'M', 'HC', '_H4_', 0, 2.5, 0.03, 0.06, 1.425, 1.25,
             0.00859824, 0.268726247]]

        expected_bonds = [
            [1, 2, 340.0, 1.09],
            [1, 3, 340.0, 1.09],
            [1, 4, 340.0, 1.09],
            [1, 5, 340.0, 1.09]]

        expected_angles = [
            [2, 1, 3, 33.0, 107.8],
            [2, 1, 4, 33.0, 107.8],
            [2, 1, 5, 33.0, 107.8],
            [3, 1, 4, 33.0, 107.8],
            [3, 1, 5, 33.0, 107.8],
            [4, 1, 5, 33.0, 107.8]]

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_nonbonding,
                         expected_bonds=expected_bonds,
                         expected_angles=expected_angles)

        # Load molecule
        opls2005 = OPLS2005ForceField()
        molecule = Molecule(get_data_file_path('ligands/ethylene.pdb'))
        ffld_file = get_data_file_path('tests/ETL_ffld_output.txt')

        # Workaround to parameterize with OPLS2005 without the Schrodinger
        # dependency
        parameters = parameterize_opls2005(opls2005, molecule, ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Define expected parameters
        expected_propers = [
            [3, 1, 2, 5, 7.0, -1, 2, 0.0],
            [3, 1, 2, 6, 7.0, -1, 2, 0.0],
            [4, 1, 2, 5, 7.0, -1, 2, 0.0],
            [4, 1, 2, 6, 7.0, -1, 2, 0.0]]

        expected_impropers = [
            [3, 4, 1, 2, 15.0, -1, 2],
            [5, 6, 2, 1, 15.0, -1, 2]]

        # Check it up
        check_parameters(topology,
                         expected_propers=expected_propers,
                         expected_impropers=expected_impropers)

    def test_openffopls2005_parameterizer(self):
        """It checks the parameterized method."""

        from peleffy.topology import Molecule
        from peleffy.topology import Topology
        from peleffy.forcefield import OpenFFOPLS2005ForceField
        from peleffy.utils import get_data_file_path
        from .utils import check_parameters, parameterize_openffopls2005

        FORCE_FIELD_NAME = 'openff_unconstrained-1.2.1.offxml'

        # Load molecule
        hybridff = OpenFFOPLS2005ForceField(FORCE_FIELD_NAME)
        molecule = Molecule(get_data_file_path('ligands/methane.pdb'))
        ffld_file = get_data_file_path('tests/MET_ffld_output.txt')

        # Workaround to skip Schrodinger dependency
        parameters = parameterize_openffopls2005(hybridff, molecule,
                                                 ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Define expected parameters
        expected_off_nonbonding = [
            [1, 0, 'M', 'OFFT', '_C1_', 0, 3.3996695084235347, 0.1094,
             -0.1088, 0, 1.6998347542117673, 0, 0],
            [2, 1, 'M', 'OFFT', '_H1_', 0, 2.649532787749369, 0.0157,
             0.0267, 0, 1.3247663938746845, 0, 0],
            [3, 1, 'M', 'OFFT', '_H2_', 0, 2.649532787749369, 0.0157,
             0.0267, 0, 1.3247663938746845, 0, 0],
            [4, 1, 'M', 'OFFT', '_H3_', 0, 2.649532787749369, 0.0157,
             0.0267, 0, 1.3247663938746845, 0, 0],
            [5, 1, 'M', 'OFFT', '_H4_', 0, 2.649532787749369, 0.0157,
             0.0267, 0, 1.3247663938746845, 0, 0]]

        expected_off_bonds = [
            [1, 2, 376.8940758588, 1.094223427522],
            [1, 3, 376.8940758588, 1.094223427522],
            [1, 4, 376.8940758588, 1.094223427522],
            [1, 5, 376.8940758588, 1.094223427522]]

        expected_off_angles = [
            [2, 1, 3, 33.78875634641, 110.2468561538],
            [2, 1, 4, 33.78875634641, 110.2468561538],
            [2, 1, 5, 33.78875634641, 110.2468561538],
            [3, 1, 4, 33.78875634641, 110.2468561538],
            [3, 1, 5, 33.78875634641, 110.2468561538],
            [4, 1, 5, 33.78875634641, 110.2468561538]]

        expected_opls_nonbonding = [
            [1, 0, 'M', 'CT', '_C1_', 0, 3.5, 0.066, -0.1088, 1.975, 1.75,
             0.005, -0.74168571],
            [2, 1, 'M', 'HC', '_H1_', 0, 2.5, 0.03, 0.0267, 1.425, 1.25,
             0.00859824, 0.268726247],
            [3, 1, 'M', 'HC', '_H2_', 0, 2.5, 0.03, 0.0267, 1.425, 1.25,
             0.00859824, 0.268726247],
            [4, 1, 'M', 'HC', '_H3_', 0, 2.5, 0.03, 0.0267, 1.425, 1.25,
             0.00859824, 0.268726247],
            [5, 1, 'M', 'HC', '_H4_', 0, 2.5, 0.03, 0.0267, 1.425, 1.25,
             0.00859824, 0.268726247]]

        expected_opls_bonds = [
            [1, 2, 340.0, 1.09],
            [1, 3, 340.0, 1.09],
            [1, 4, 340.0, 1.09],
            [1, 5, 340.0, 1.09]]

        expected_opls_angles = [
            [2, 1, 3, 33.0, 107.8],
            [2, 1, 4, 33.0, 107.8],
            [2, 1, 5, 33.0, 107.8],
            [3, 1, 4, 33.0, 107.8],
            [3, 1, 5, 33.0, 107.8],
            [4, 1, 5, 33.0, 107.8]]

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_off_nonbonding,
                         expected_bonds=expected_off_bonds,
                         expected_angles=expected_off_angles)

        # Hybrid force field set up
        hybridff.set_nonbonding_parameters('OPLS2005')

        # Workaround to skip Schrodinger dependency
        parameters = parameterize_openffopls2005(hybridff, molecule,
                                                 ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_opls_nonbonding,
                         expected_bonds=expected_off_bonds,
                         expected_angles=expected_off_angles)

        # Hybrid force field set up
        hybridff.set_nonbonding_parameters('OpenFF')
        hybridff.set_bond_parameters('OPLS2005')

        # Workaround to skip Schrodinger dependency
        parameters = parameterize_openffopls2005(hybridff, molecule,
                                                 ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_off_nonbonding,
                         expected_bonds=expected_opls_bonds,
                         expected_angles=expected_off_angles)

        # Hybrid force field set up
        hybridff.set_nonbonding_parameters('OpenFF')
        hybridff.set_bond_parameters('OpenFF')
        hybridff.set_angle_parameters('OPLS2005')

        # Workaround to skip Schrodinger dependency
        parameters = parameterize_openffopls2005(hybridff, molecule,
                                                 ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_off_nonbonding,
                         expected_bonds=expected_off_bonds,
                         expected_angles=expected_opls_angles)

        # Hybrid force field set up
        hybridff.set_nonbonding_parameters('OPLS2005')
        hybridff.set_bond_parameters('OPLS2005')
        hybridff.set_angle_parameters('OPLS2005')

        # Workaround to skip Schrodinger dependency
        parameters = parameterize_openffopls2005(hybridff, molecule,
                                                 ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_opls_nonbonding,
                         expected_bonds=expected_opls_bonds,
                         expected_angles=expected_opls_angles)

        # Load molecule
        hybridff = OpenFFOPLS2005ForceField(FORCE_FIELD_NAME)
        molecule = Molecule(get_data_file_path('ligands/ethylene.pdb'))
        ffld_file = get_data_file_path('tests/ETL_ffld_output.txt')

        # Workaround to skip Schrodinger dependency
        parameters = parameterize_openffopls2005(hybridff, molecule,
                                                 ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Define expected parameters
        expected_off_nonbonding = [
            [1, 0, 'M', 'OFFT', '_C1_', 0, 3.3996695084235347, 0.086,
             -0.218, 0, 1.6998347542117673, 0, 0],
            [2, 1, 'M', 'OFFT', '_C2_', 0, 3.3996695084235347, 0.086,
             -0.218, 0, 1.6998347542117673, 0, 0],
            [3, 1, 'M', 'OFFT', '_H1_', 0, 2.59964245953351, 0.015,
             0.109, 0, 1.299821229766755, 0, 0],
            [4, 1, 'M', 'OFFT', '_H2_', 0, 2.59964245953351, 0.015,
             0.109, 0, 1.299821229766755, 0, 0],
            [5, 2, 'M', 'OFFT', '_H3_', 0, 2.59964245953351, 0.015,
             0.109, 0, 1.299821229766755, 0, 0],
            [6, 2, 'M', 'OFFT', '_H4_', 0, 2.59964245953351, 0.015,
             0.109, 0, 1.299821229766755, 0, 0]]

        expected_off_bonds = [[1, 2, 404.83941263565, 1.372230219368],
                              [1, 3, 404.20804685, 1.085503378387],
                              [1, 4, 404.20804685, 1.085503378387],
                              [2, 5, 404.20804685, 1.085503378387],
                              [2, 6, 404.20804685, 1.085503378387]]

        expected_off_angles = [[1, 2, 5, 34.202963712735, 133.1339832262],
                               [1, 2, 6, 34.202963712735, 133.1339832262],
                               [2, 1, 3, 34.202963712735, 133.1339832262],
                               [2, 1, 4, 34.202963712735, 133.1339832262],
                               [3, 1, 4, 27.86109944498, 134.0642036482],
                               [5, 2, 6, 27.86109944498, 134.0642036482]]

        expected_off_propers = [
            [3, 1, 2, 5, 5.376019778605, -1, 2, 0.0],
            [3, 1, 2, 6, 5.376019778605, -1, 2, 0.0],
            [4, 1, 2, 5, 5.376019778605, -1, 2, 0.0],
            [4, 1, 2, 6, 5.376019778605, -1, 2, 0.0]]

        expected_off_impropers = [
            [1, 2, 5, 6, 1.1, -1, 2],
            [2, 1, 3, 4, 1.1, -1, 2]]

        expected_opls_propers = [
            [3, 1, 2, 5, 7.0, -1, 2, 0.0],
            [3, 1, 2, 6, 7.0, -1, 2, 0.0],
            [4, 1, 2, 5, 7.0, -1, 2, 0.0],
            [4, 1, 2, 6, 7.0, -1, 2, 0.0]]

        expected_opls_impropers = [
            [3, 4, 1, 2, 15.0, -1, 2],
            [5, 6, 2, 1, 15.0, -1, 2]]

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_off_nonbonding,
                         expected_bonds=expected_off_bonds,
                         expected_angles=expected_off_angles,
                         expected_propers=expected_off_propers,
                         expected_impropers=expected_off_impropers)

        # Hybrid force field set up
        hybridff.set_nonbonding_parameters('OpenFF')
        hybridff.set_dihedral_parameters('OPLS2005')

        # Workaround to skip Schrodinger dependency
        parameters = parameterize_openffopls2005(hybridff, molecule,
                                                 ffld_file)

        # Generate molecular topology
        topology = Topology(molecule, parameters)

        # Check it up
        check_parameters(topology,
                         expected_nonbonding=expected_off_nonbonding,
                         expected_bonds=expected_off_bonds,
                         expected_angles=expected_off_angles,
                         expected_propers=expected_opls_propers,
                         expected_impropers=expected_opls_impropers)
